__all__ = [
    "StationaryOnGroundUser",
    "User",
    "load_users_from_toml_file",
    "parse_user_item",
]
import tomllib
from abc import ABC, abstractmethod
from collections.abc import Hashable, Mapping
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Self

import numpy as np
from typing_extensions import deprecated

from .node import Node
from .terminal import CommunicationTerminal


@dataclass(frozen=True, kw_only=True, slots=True)
class User[T: Hashable](Node[T], ABC):
    """Base model for a user."""

    id: T
    terminals: list[CommunicationTerminal[Hashable]] = field(default_factory=list)

    @classmethod
    @abstractmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_user_item(cls, item: Mapping[str, Any]) -> Self:
        """Parse a user item."""
        ...


@dataclass(frozen=True, kw_only=True, slots=True)
class StationaryOnGroundUser[T: Hashable](User[T]):
    """Model for a stationary user on the ground."""

    id: T
    latitude: float = field(compare=False)
    longitude: float = field(compare=False)
    altitude: float = field(default=0.0, compare=False)
    minimum_elevation: float = field(compare=False)
    terminals: list[CommunicationTerminal[Hashable]] = field(default_factory=list, compare=False)

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_user_item(cls, item: Mapping[str, Any]) -> Self:
        return cls(
            id=item["id"],
            latitude=np.deg2rad(item["lat_deg"]),
            longitude=np.deg2rad(item["lon_deg"]),
            altitude=item["alt_m"],
            minimum_elevation=np.deg2rad(item["min_el_deg"]),
        )


_USER_TYPES: dict[str, type[User[Hashable]]] = {cls.class_name(): cls for cls in (StationaryOnGroundUser[Hashable],)}


@deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
def parse_user_item(item: Mapping[str, Any]) -> User[Hashable]:
    """Parse a user item."""
    user_type = item["type"]
    return _USER_TYPES[user_type].parse_user_item(item)


@deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
def load_users_from_toml_file(toml_file_path: str | Path) -> list[User[Hashable]]:
    """Load users from a TOML file."""
    toml_file_path = Path(toml_file_path)
    with toml_file_path.open("rb") as f:
        toml_data = tomllib.load(f)
    return list(map(parse_user_item, toml_data["users"]))



__all__ = [
    "GatewayToInternetTopologyBuilder",
]

from collections.abc import Collection

import networkx as nx

from cosmica.models import Gateway, Internet


class GatewayToInternetTopologyBuilder:
    def build(
        self,
        *,
        gateways: Collection[Gateway],
        internet: Internet,
    ) -> nx.Graph:
        graph = nx.Graph()
        graph.add_nodes_from(gateways)
        graph.add_node(internet)
        graph.add_edges_from((gateway, internet) for gateway in gateways)

        return graph



__all__ = [
    "CircularSatelliteOrbitModel",
    "EllipticalSatelliteOrbitModel",
    "GravityModel",
]
from enum import Enum

import numpy as np
from pydantic import BaseModel, ConfigDict
from sgp4.model import WGS72, WGS72OLD, WGS84


class GravityModel(Enum):
    WGS72 = WGS72
    WGS72OLD = WGS72OLD
    WGS84 = WGS84


class SatelliteOrbitModel(BaseModel):
    """Base model for a satellite orbit."""


class CircularSatelliteOrbitModel(SatelliteOrbitModel):
    # arbitrary_types_allowed is required to allow numpy.datetime64
    model_config = ConfigDict(arbitrary_types_allowed=True)

    semi_major_axis: float
    inclination: float
    raan: float
    phase_at_epoch: float
    epoch: np.datetime64


class EllipticalSatelliteOrbitModel(SatelliteOrbitModel):
    # arbitrary_types_allowed is required to allow numpy.datetime64
    model_config = ConfigDict(arbitrary_types_allowed=True)

    semi_major_axis: float
    inclination: float
    raan: float
    phase_at_epoch: float
    epoch: np.datetime64
    satnum: int  # Satellite number
    gravity_model: GravityModel = GravityModel.WGS84
    drag_coeff: float  # B star drag coeffient given in [1/earth radii]
    eccentricity: float
    argpo: float  # Argument of Perigee



__all__ = [
    "BOLTZ_CONST",
    "EARTH_MU",
    "EARTH_RADIUS",
    "EARTH_ROTATION_RATE",
]
import math
from typing import Final

# Earth gravitational constant from GRS80 model: 3.986005e14 m3/s2.
EARTH_MU: Final[float] = 3.986005e14  # [m3/s2]

# Earth equatorial radius as defined by IAU 2015 resolution B3: 6.3781e6 (m).
EARTH_RADIUS: Final[float] = 6.378137e6  # [m]

# Earth rotation rate
EARTH_ROTATION_RATE: Final[float] = 2 * math.pi / 86164.0905  # [rad/s]

SPEED_OF_LIGHT: Final[float] = 299_792_458.0  # [m/s]

# Boltzmann Constant
BOLTZ_CONST: Final[float] = 1.380649e-23  # [J/K]



__all__ = [
    "CircularSatelliteOrbit",
    "CircularSatelliteOrbitPropagator",
    "EllipticalSatelliteOrbit",
    "EllipticalSatelliteOrbitPropagator",
    "MOPCSatelliteKey",
    "MultiOrbitalPlaneConstellation",
    "SatelliteConstellation",
    "SatelliteOrbit",
    "SatelliteOrbitState",
    "get_sun_direction_eci",
    "make_satellite_orbit",
]
from .constellation import MOPCSatelliteKey, MultiOrbitalPlaneConstellation, SatelliteConstellation
from .orbit import (
    CircularSatelliteOrbit,
    CircularSatelliteOrbitPropagator,
    EllipticalSatelliteOrbit,
    EllipticalSatelliteOrbitPropagator,
    SatelliteOrbit,
    SatelliteOrbitState,
    make_satellite_orbit,
)
from .sun_dynamics import get_sun_direction_eci
# We don't export from the `plotting` module. Maybe we should move it to `utils`?



__all__ = [
    "ForwardingTableTimeList",
]
from dataclasses import dataclass, field

import numpy as np

from cosmica.models.node import Node

from .forwarding_table import ForwardingTable


@dataclass(kw_only=True, slots=True)
class ForwardingTableTimeList:
    time_for_snapshots: list[np.datetime64]
    nominal_forwarding_table_for_snapshots: list[ForwardingTable]
    backup_forwarding_tables_for_snapshots: list[dict[frozenset[frozenset[Node]], ForwardingTable]] = field(
        default_factory=list,
    )

    # TODO(Takashima): ある時刻以降のforwarding_table_time_listを更新



__all__ = [
    "CommData",
    "CommDataDemand",
    "CommDataLSA",
]

from collections.abc import Hashable
from dataclasses import dataclass

import numpy as np

from cosmica.models.node import Node


@dataclass(kw_only=True, slots=True)
class CommData:
    """Base model for a communication data."""

    data_size: int
    packet_size: int = 10000
    packet_num: int
    dst_node: Node
    next_node: Node
    current_position: Node | tuple[Node, Node]
    path: list[Node]
    generated_time: np.datetime64
    time: np.datetime64
    time_from_generated: float
    time_remaining_for_current_position: float
    delay: float = np.nan
    reach_dst: bool = False
    packet_loss: bool = False


@dataclass(kw_only=True, slots=True)
class CommDataDemand(CommData):
    """Model for a communication data for demand."""

    demand_id: Hashable


@dataclass(kw_only=True, slots=True)
class CommDataLSA(CommData):
    """Model for a communication data for LSA."""

    failure_position: Node | tuple[Node, Node]



from __future__ import annotations

__all__ = [
    "DynamicsData",
]

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt

from cosmica.models import Satellite


def _check_item_shape_if_any[T: Satellite](
    data: dict[T, npt.NDArray[np.floating]],
    target_shape: tuple[int, ...],
    slice_: int | slice | None = None,
) -> bool:
    try:
        slice_ = slice_ or slice(None)
        return next(iter(data.values())).shape[slice_] == target_shape
    except StopIteration:
        return True


@dataclass(frozen=True, kw_only=True, slots=True)
class DynamicsData[T: Satellite]:
    time: npt.NDArray[np.datetime64]
    dcm_eci2ecef: npt.NDArray[np.floating]
    satellite_position_eci: dict[T, npt.NDArray[np.floating]]
    satellite_velocity_eci: dict[T, npt.NDArray[np.floating]]
    satellite_position_ecef: dict[T, npt.NDArray[np.floating]]
    satellite_attitude_angular_velocity_eci: dict[T, npt.NDArray[np.floating]]
    sun_direction_eci: npt.NDArray[np.floating]
    sun_direction_ecef: npt.NDArray[np.floating]

    @property
    def data_shape(self) -> tuple[int, ...]:
        return self.time.shape

    def __post_init__(self) -> None:
        data_shape = self.time.shape
        assert self.dcm_eci2ecef.shape[:-2] == data_shape
        assert _check_item_shape_if_any(self.satellite_position_eci, data_shape, slice(-1))
        assert _check_item_shape_if_any(self.satellite_velocity_eci, data_shape, slice(-1))
        assert _check_item_shape_if_any(self.satellite_position_ecef, data_shape, slice(-1))
        assert _check_item_shape_if_any(self.satellite_attitude_angular_velocity_eci, data_shape, slice(-1))
        assert self.sun_direction_eci.shape[:-1] == data_shape
        assert self.sun_direction_ecef.shape[:-1] == data_shape

    def __getitem__(self, item: int | slice) -> DynamicsData[T]:
        return DynamicsData(
            time=self.time[item],
            dcm_eci2ecef=self.dcm_eci2ecef[item],
            satellite_position_eci={key: value[item] for key, value in self.satellite_position_eci.items()},
            satellite_velocity_eci={key: value[item] for key, value in self.satellite_velocity_eci.items()},
            satellite_position_ecef={key: value[item] for key, value in self.satellite_position_ecef.items()},
            satellite_attitude_angular_velocity_eci={
                key: value[item] for key, value in self.satellite_attitude_angular_velocity_eci.items()
            },
            sun_direction_eci=self.sun_direction_eci[item],
            sun_direction_ecef=self.sun_direction_ecef[item],
        )



__all__ = [
    "ElevationBasedG2CTopologyBuilder",
    "GroundToConstellationTopologyBuilder",
    "ManualG2CTopologyBuilder",
]
import logging
from abc import ABC, abstractmethod
from collections.abc import Collection
from itertools import product

import networkx as nx
import numpy as np
import numpy.typing as npt

from cosmica.dtos import DynamicsData
from cosmica.dynamics import SatelliteConstellation
from cosmica.models import ConstellationSatellite, Gateway, Node, StationaryOnGroundUser
from cosmica.utils.coordinates import ecef2aer

logger = logging.getLogger(__name__)


class GroundToConstellationTopologyBuilder[TConstellation: SatelliteConstellation, TNode: Node, TGraph: nx.Graph](ABC):
    @abstractmethod
    def build(
        self,
        *,
        constellation: TConstellation,
        ground_nodes: Collection[TNode],
        dynamics_data: DynamicsData,
    ) -> list[TGraph]: ...


class ElevationBasedG2CTopologyBuilder(
    GroundToConstellationTopologyBuilder[SatelliteConstellation, Gateway | StationaryOnGroundUser, nx.Graph],
):
    def build(
        self,
        *,
        constellation: SatelliteConstellation,
        ground_nodes: Collection[Gateway | StationaryOnGroundUser],
        dynamics_data: DynamicsData,
    ) -> list[nx.Graph]:
        logger.info(f"Number of time steps: {len(dynamics_data.time)}")
        ground_nodes = list(ground_nodes)

        n_satellites = len(dynamics_data.satellite_position_eci)
        visibility = np.zeros((len(ground_nodes), n_satellites, len(dynamics_data.time)), dtype=np.bool_)
        for (ground_node_idx, ground_node), (sat_idx, satellite) in product(
            enumerate(ground_nodes),
            enumerate(constellation.satellites),
        ):
            _, elevation, _ = ecef2aer(
                x=dynamics_data.satellite_position_ecef[satellite][:, 0],
                y=dynamics_data.satellite_position_ecef[satellite][:, 1],
                z=dynamics_data.satellite_position_ecef[satellite][:, 2],
                lat0=ground_node.latitude,
                lon0=ground_node.longitude,
                h0=ground_node.altitude,
                deg=False,
            )
            visibility[ground_node_idx, sat_idx, :] = elevation >= ground_node.minimum_elevation

        def construct_graph(visibility: npt.NDArray[np.bool_]) -> nx.Graph:
            graph = nx.Graph()
            graph.add_nodes_from(constellation.satellites)
            graph.add_nodes_from(ground_nodes)

            for (ground_node_idx, ground_node), (sat_idx, satellite) in product(
                enumerate(ground_nodes),
                enumerate(constellation.satellites),
            ):
                if visibility[ground_node_idx, sat_idx]:
                    graph.add_edge(ground_node, satellite)

            return graph

        return [construct_graph(visibility[:, :, time_idx]) for time_idx in range(len(dynamics_data.time))]


class ManualG2CTopologyBuilder(
    GroundToConstellationTopologyBuilder[SatelliteConstellation, Gateway | StationaryOnGroundUser, nx.Graph],
):
    def __init__(self, custom_connections: dict[Gateway | StationaryOnGroundUser, ConstellationSatellite]) -> None:
        self.custom_connections: dict[Gateway | StationaryOnGroundUser, ConstellationSatellite] = custom_connections

    def build(
        self,
        *,
        constellation: SatelliteConstellation,
        ground_nodes: Collection[Gateway | StationaryOnGroundUser],
        dynamics_data: DynamicsData,
    ) -> list[nx.Graph]:
        logger.info(f"Number of time steps: {len(dynamics_data.time)}")
        ground_nodes = list(ground_nodes)

        def construct_graph() -> nx.Graph:
            graph = nx.Graph()
            graph.add_nodes_from(constellation.satellites)
            graph.add_nodes_from(ground_nodes)

            for ground_node, satellite in self.custom_connections.items():
                graph.add_edge(ground_node, satellite)

            return graph

        # dynamics_data.time の長さに応じたグラフを返す
        return [construct_graph() for _ in range(len(dynamics_data.time))]


# class MaxElevationHandOverG2CTopologyBuilder(
#     GroundToConstellationTopologyBuilder[SatelliteConstellation, Gateway | StationaryOnGroundUser, nx.Graph],
# ):
#     def build(
#         self,
#         *,
#         constellation: SatelliteConstellation,
#         ground_nodes: Collection[Gateway | StationaryOnGroundUser],
#         dynamics_data: DynamicsData,
#     ) -> list[nx.Graph]:
#         logger.info(f"Number of time steps: {len(dynamics_data.time)}")
#         ground_nodes = list(ground_nodes)

#         n_satellites = len(dynamics_data.satellite_position_eci)
#         visibility = np.zeros((len(ground_nodes), n_satellites, len(dynamics_data.time)), dtype=np.bool_)
#         elevation_angles = np.zeros((len(ground_nodes), n_satellites, len(dynamics_data.time)))
#         for (ground_node_idx, ground_node), (sat_idx, satellite) in product(
#             enumerate(ground_nodes),
#             enumerate(constellation.satellites),
#         ):
#             _, elevation, _ = ecef2aer(
#                 x=dynamics_data.satellite_position_ecef[satellite][:, 0],
#                 y=dynamics_data.satellite_position_ecef[satellite][:, 1],
#                 z=dynamics_data.satellite_position_ecef[satellite][:, 2],
#                 lat0=ground_node.latitude,
#                 lon0=ground_node.longitude,
#                 h0=ground_node.altitude,
#                 deg=False,
#             )
#             elevation_angles[ground_node_idx, sat_idx, :] = elevation
#             visibility[ground_node_idx, sat_idx, :] = elevation >= ground_node.minimum_elevation

#         def construct_graph(visibility: npt.NDArray[np.bool_]) -> nx.Graph:
#             graph = nx.Graph()
#             graph.add_nodes_from(constellation.satellites)
#             graph.add_nodes_from(ground_nodes)

#             for (ground_node_idx, ground_node), (sat_idx, satellite) in product(
#                 enumerate(ground_nodes),
#                 enumerate(constellation.satellites),
#             ):
#                 if visibility[ground_node_idx, sat_idx]:
#                     graph.add_edge(ground_node, satellite)

#             return graph

#         return [construct_graph(visibility[:, :, time_idx]) for time_idx in range(len(dynamics_data.time))]


class MaxVisibilityHandOverG2CTopologyBuilder(
    GroundToConstellationTopologyBuilder[SatelliteConstellation, Gateway | StationaryOnGroundUser, nx.Graph],
):
    def build(  # noqa: C901
        self,
        *,
        constellation: SatelliteConstellation,
        ground_nodes: Collection[Gateway | StationaryOnGroundUser],
        dynamics_data: DynamicsData,
    ) -> list[nx.Graph]:
        logger.info(f"Number of time steps: {len(dynamics_data.time)}")
        ground_nodes = list(ground_nodes)

        n_time = len(dynamics_data.time)
        n_gateways = len(ground_nodes)
        n_satellites = len(dynamics_data.satellite_position_eci)
        visibility = np.zeros((n_gateways, n_satellites, n_time), dtype=np.bool_)
        for (ground_node_idx, ground_node), (sat_idx, satellite) in product(
            enumerate(ground_nodes),
            enumerate(constellation.satellites),
        ):
            _, elevation, _ = ecef2aer(
                x=dynamics_data.satellite_position_ecef[satellite][:, 0],
                y=dynamics_data.satellite_position_ecef[satellite][:, 1],
                z=dynamics_data.satellite_position_ecef[satellite][:, 2],
                lat0=ground_node.latitude,
                lon0=ground_node.longitude,
                h0=ground_node.altitude,
                deg=False,
            )
            visibility[ground_node_idx, sat_idx, :] = elevation >= ground_node.minimum_elevation

        # 残接続時間を計算
        left_visibility_time_step = np.zeros((n_gateways, n_satellites, n_time), dtype=np.int_)
        for time_idx in reversed(range(n_time)):
            left_visibility_time_step[:, :, time_idx] = (
                np.where(
                    visibility[:, :, time_idx],
                    left_visibility_time_step[:, :, time_idx + 1] + 1,
                    0,
                )
                if time_idx != n_time - 1
                else visibility[:, :, time_idx].astype(int)
            )

        # 地上局に対する通信する衛星の選択
        select_sat_idx = np.full(
            n_gateways,
            -1,
            dtype=np.int_,
        )  # -1: 通信しない TODO: -1でなくNanしたいが, int型ではエラーが出る
        # connection_accumulation_time = np.zeros(n_gateways, dtype=np.float64)
        link_available = np.zeros((n_gateways, n_satellites, n_time), dtype=np.bool_)

        def select_max_visibility_satellite(
            ground_node_idx: int,
            time_idx: int,
        ) -> None:
            # 他の地上局と通信している衛星は除く
            masked_left_visibility_time_step = np.copy(left_visibility_time_step[ground_node_idx, :, time_idx])
            for sat_idx in select_sat_idx:
                if sat_idx == -1:
                    continue
                masked_left_visibility_time_step[sat_idx] = 0
            # 残り可視時間が最大の衛星を選択
            select_sat_idx[ground_node_idx] = np.argmax(masked_left_visibility_time_step)

        for time_idx in range(n_time):
            for ground_node_idx, _ground_node in enumerate(ground_nodes):
                if time_idx == 0:
                    # すべての衛星に対して可視でない
                    if left_visibility_time_step[ground_node_idx, :, time_idx].sum == 0:
                        select_sat_idx[ground_node_idx] = -1
                        continue
                    # 残り可視時間が最大の衛星を選択
                    select_max_visibility_satellite(ground_node_idx, time_idx)
                else:
                    # すべての衛星に対して可視でない
                    if left_visibility_time_step[ground_node_idx, :, time_idx].sum == 0:
                        select_sat_idx[ground_node_idx] = -1
                        continue
                    # 前のタイムステップで接続している衛星がいる
                    if select_sat_idx[ground_node_idx] >= 0:
                        # 現在のタイムステップも可視 → 同じ衛星に接続
                        if left_visibility_time_step[ground_node_idx, select_sat_idx[ground_node_idx], time_idx] > 0:
                            # # 累積連続可視時間を更新
                            # connection_accumulation_time[ground_node_idx] += (
                            #     dynamics_data.time[time_idx] - dynamics_data.time[time_idx - 1]
                            # ) / np.timedelta64(1, "s")
                            pass
                        else:
                            # 残り可視時間が最大の衛星を選択
                            select_max_visibility_satellite(ground_node_idx, time_idx)
                    # 前のタイムステップで接続している衛星がいない
                    else:
                        # 残り可視時間が最大の衛星を選択
                        select_max_visibility_satellite(ground_node_idx, time_idx)
                link_available[ground_node_idx, select_sat_idx[ground_node_idx], time_idx] = True

        def construct_graph(link_available: npt.NDArray[np.bool_]) -> nx.Graph:
            graph = nx.Graph()
            graph.add_nodes_from(constellation.satellites)
            graph.add_nodes_from(ground_nodes)

            for (ground_node_idx, ground_node), (sat_idx, satellite) in product(
                enumerate(ground_nodes),
                enumerate(constellation.satellites),
            ):
                if link_available[ground_node_idx, sat_idx]:
                    graph.add_edge(ground_node, satellite)

            return graph

        return [construct_graph(link_available[:, :, time_idx]) for time_idx in range(len(dynamics_data.time))]



from __future__ import annotations

__all__ = [
    "visualize_multi_orbital_plane_constellation",
]
from typing import TYPE_CHECKING

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

if TYPE_CHECKING:
    from collections.abc import Mapping

    import numpy.typing as npt
    from mpl_toolkits.mplot3d.axes3d import Axes3D

    from cosmica.models import ConstellationSatellite

if TYPE_CHECKING:
    from cosmica.dynamics import (
        CircularSatelliteOrbit,
        MultiOrbitalPlaneConstellation,
        SatelliteOrbitState,
    )

from cosmica.utils.constants import EARTH_RADIUS


def visualize_multi_orbital_plane_constellation(
    constellation: MultiOrbitalPlaneConstellation[CircularSatelliteOrbit],
    propagation_result: Mapping[ConstellationSatellite, SatelliteOrbitState],
    time_index: int = 0,
) -> None:
    fig = plt.figure()
    ax: Axes3D = fig.add_subplot(111, projection="3d")

    cmap = mpl.colormaps["tab20"]

    # Plot the Earth
    ax.plot_surface(*_ms(0, 0, 0, EARTH_RADIUS), color="blue", alpha=0.2)

    # Plot the orbit trajectory of the first satellite in each orbital plane
    for i, plane_id in enumerate(constellation.plane_ids):
        satellite = constellation.plane_id_to_satellites[plane_id][0]
        ax.plot(
            propagation_result[satellite].position_eci[:, 0],
            propagation_result[satellite].position_eci[:, 1],
            propagation_result[satellite].position_eci[:, 2],
            color=cmap(i),
            linewidth=0.5,
            label=f"Plane {plane_id}",
        )
        for satellite in constellation.satellites:
            # Plot the position of each satellite at the `time_index`
            ax.plot(
                propagation_result[satellite].position_eci[time_index, 0],
                propagation_result[satellite].position_eci[time_index, 1],
                propagation_result[satellite].position_eci[time_index, 2],
                "ro",
                markersize=2,
            )

    # Set plot labels and aspect ratio
    ax.set_xlabel("X (km)")
    ax.set_ylabel("Y (km)")
    ax.set_zlabel("Z (km)")
    ax.set_box_aspect([1, 1, 1])

    ax.legend()

    plt.show()


def _ms(
    x: float,
    y: float,
    z: float,
    radius: float,
    resolution: float = 20,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """Return the coordinates for plotting a sphere centered at (x,y,z)."""
    u, v = np.mgrid[0 : 2 * np.pi : resolution * 2j, 0 : np.pi : resolution * 1j]  # type: ignore[misc]
    xx = radius * np.cos(u) * np.sin(v) + x
    yy = radius * np.sin(u) * np.sin(v) + y
    zz = radius * np.cos(v) + z
    return (xx, yy, zz)



__all__ = [
    "CommLinkPerformance",
    "MemorylessCommLinkCalculator",
    "MemorylessCommLinkCalculatorWrapper",
]
from abc import ABC, abstractmethod
from collections.abc import Collection, Sequence
from typing import TypedDict

import numpy as np

from cosmica.dtos import DynamicsData
from cosmica.models import Node


# Use TypedDict instead of dataclass so it can be treated nicely by nx.set_edge_attributes
# Ref: https://networkx.org/documentation/stable/reference/generated/networkx.classes.function.set_edge_attributes.html
class CommLinkPerformance(TypedDict):
    link_capacity: float  # [bps]
    delay: float  # [s]
    link_available: bool


class MemorylessCommLinkCalculator[T: Node, U: Node](ABC):
    @abstractmethod
    def calc(
        self,
        edges: Collection[tuple[T, U]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,
    ) -> dict[tuple[T, U], CommLinkPerformance]:
        """Calculate communication link performance for each edge in a network."""
        raise NotImplementedError


class CommLinkCalculator[T: Node, U: Node](ABC):
    @abstractmethod
    def calc(
        self,
        edges_time_series: Sequence[Collection[tuple[T, U]]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,
    ) -> list[dict[tuple[T, U], CommLinkPerformance]]:
        """Calculate communication link performance for each edge in a network."""
        raise NotImplementedError


class MemorylessCommLinkCalculatorWrapper[T: Node, U: Node](CommLinkCalculator[T, U]):
    """Convert a memoryless calculator to a time series calculator by calling the memoryless one at each time step."""

    def __init__(
        self,
        memoryless_calculator: MemorylessCommLinkCalculator[T, U],
        /,
    ) -> None:
        self.memoryless_calculator = memoryless_calculator

    def calc(
        self,
        edges_time_series: Sequence[Collection[tuple[T, U]]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,
    ) -> list[dict[tuple[T, U], CommLinkPerformance]]:
        assert len(edges_time_series) == len(
            dynamics_data.time,
        ), "edges_time_series must have the same length as dynamics_data.time"
        return [
            self.memoryless_calculator.calc(edges=edges, dynamics_data=dynamics_data[time_index], rng=rng)
            for time_index, edges in enumerate(edges_time_series)
        ]



__all__ = [
    "GatewayToGatewayTopologyBuilder",
]

from collections.abc import Collection, Hashable

import networkx as nx

from cosmica.models import Gateway


class GatewayToGatewayTopologyBuilder:
    def build(
        self,
        *,
        gateways: Collection[Gateway[Hashable]],
    ) -> nx.Graph:
        graph = nx.Graph()
        gateway_list = list(gateways)
        graph.add_nodes_from(gateway_list)

        # すべてのgateway同士を接続する(完全グラフ)
        for i, gateway1 in enumerate(gateway_list):
            for gateway2 in gateway_list[i + 1 :]:
                graph.add_edge(gateway1, gateway2)

        return graph



import warnings
from typing import TYPE_CHECKING

import numpy as np

from cosmica.models.node import Node

from .node_knowledge import NodeKnowledge

if TYPE_CHECKING:
    from networkx import Graph

    from cosmica.experimental_packet_routing.forwarding_table import ForwardingTableInformation

    from .forwarding_table import ForwardingTable


def calc_next_node_from_node_knowledge(  # noqa: C901, PLR0912, PLR0915
    node_knowledge: NodeKnowledge,
    dst_node: Node,
    path: list[Node],
    current_time: np.datetime64,
    *,
    enable_random_routing_when_edge_failure: bool = False,
    prevent_loop: bool = False,
) -> Node | None:
    """Calculate the next node from the network information."""
    rng: np.random.Generator = np.random.default_rng()

    # Step 1: Find the index of the closest time snapshot before or at current_time
    time_index = max(
        i for i, time in enumerate(node_knowledge.forwarding_table_time_list.time_for_snapshots) if time <= current_time
    )

    # Step 2: Select the forwarding table
    do_backup: bool = False

    # Use the nominal forwarding table for the identified time snapshot
    forwarding_table: ForwardingTable = (
        node_knowledge.forwarding_table_time_list.nominal_forwarding_table_for_snapshots[time_index]
    )

    # TODO(Takashima): 3つ以上のリンク切断バックアップの処理を検討する

    # Backup forwarding tables (both single and dual) are used when the routing recalculation period is active
    if node_knowledge.forwarding_table_time_list.backup_forwarding_tables_for_snapshots:
        backup_forwarding_tables: dict[frozenset[frozenset[Node]], ForwardingTable] = (
            node_knowledge.forwarding_table_time_list.backup_forwarding_tables_for_snapshots[time_index]
        )
        if any(start <= current_time <= end for start, end in node_knowledge.routing_recalculation_period):
            # Dual link backup forwarding tables
            dual_backup_forwarding_table_candidates: dict[frozenset[frozenset[Node]], ForwardingTable] = {}

            for failure_assumed_edges, backup_table in backup_forwarding_tables.items():
                if len(failure_assumed_edges) == 2:  # Dual backup case
                    # frozensetを使用しているので、集合演算で直接比較可能
                    failure_edge_frozensets = {frozenset([u, v]) for u, v in node_knowledge.failure_assumed_edge_list}
                    if failure_assumed_edges.issubset(failure_edge_frozensets):
                        dual_backup_forwarding_table_candidates[failure_assumed_edges] = backup_table

            # TODO(Takashima): 候補が複数ある場合の処理については、もう少し検討する
            if len(dual_backup_forwarding_table_candidates) == 0:
                pass
            elif len(dual_backup_forwarding_table_candidates) == 1:
                forwarding_table = next(iter(dual_backup_forwarding_table_candidates.values()))
                do_backup = True
            elif len(dual_backup_forwarding_table_candidates) > 1:
                pass

        # Single link backup forwarding tables are used when the routing recalculation period is active
        if any(start <= current_time <= end for start, end in node_knowledge.routing_recalculation_period):
            # TODO(Takashima): 複数の同時リンク切断への対応方法は色々あるので検討する
            single_backup_forwarding_table_candidates: dict[frozenset[frozenset[Node]], ForwardingTable] = {}
            # TODO(Takashima): エッジを探すときに、frozensetを使用しているため順序を気にする必要はない
            for edge_key, backup_table in backup_forwarding_tables.items():
                if len(edge_key) == 1:  # Single edge failure
                    edge_set = next(iter(edge_key))  # Get the single edge frozenset
                    # frozensetなので順序を気にせずに直接比較可能
                    failure_edge_frozensets = {frozenset([u, v]) for u, v in node_knowledge.failure_assumed_edge_list}
                    if edge_set in failure_edge_frozensets:
                        single_backup_forwarding_table_candidates[edge_key] = backup_table

            # TODO(Takashima): 候補が複数ある場合の処理については、もう少し検討する
            if len(single_backup_forwarding_table_candidates) == 0:
                pass
            elif len(single_backup_forwarding_table_candidates) == 1:
                forwarding_table = next(iter(single_backup_forwarding_table_candidates.values()))
                do_backup = True
            elif len(single_backup_forwarding_table_candidates) > 1:
                # # すべてのバックアップフォワーディングテーブルの次宛先が同じ場合は、バックアップフォワーディングテーブルのいずれかを選択する  # noqa: E501
                # if all(
                #     forwarding_table_candidate.find_entry(dst_node).next_node
                #     == next(iter(forwarding_table_candidates.values())).find_entry(dst_node).next_node
                #     for forwarding_table_candidate in forwarding_table_candidates.values()
                # ):
                #     forwarding_table = next(iter(forwarding_table_candidates.values()))
                #     do_backup = True

                # # (u,v)が地上局-コンステ間のエッジの場合は、そのエントリを選択する
                # for (u, v), forwarding_table_candidate in forwarding_table_candidates.items():
                #     if (isinstance(u, Gateway) and isinstance(v, ConstellationSatellite)) or (
                #         isinstance(v, Gateway) and isinstance(u, ConstellationSatellite)
                #     ):
                #         forwarding_table = forwarding_table_candidate
                #         do_backup = True
                #         break
                pass

        # ループを防ぐために、直前に通ったエッジのバックアップテーブルを保持している場合は、そのエッジのバックアップフォワーディングテーブルを使用する  # noqa: E501
        # ただしバックアップフォワーディングテーブルを使用する場合は、そのバックアップフォワーディングテーブルに従う
        if prevent_loop and (not do_backup) and (len(path) >= 2):
            # frozensetを使用しているので順序を気にせずに直接検索可能
            edge_key = frozenset([frozenset([path[-2], path[-1]])])

            if edge_key in backup_forwarding_tables:
                forwarding_table = backup_forwarding_tables[edge_key]

    # Step 3: Find the forwarding entry for the destination node
    forwarding_entry: ForwardingTableInformation | None = forwarding_table.find_entry(dst_node)

    if forwarding_entry:
        # Step 4: Return the next node for the destination
        # If random routing is enabled, return a random next node from the list of next nodes
        next_node: Node | None = forwarding_entry.next_node
    else:
        # If no forwarding entry is found, raise an exception or return random next node
        msg: str = f"No forwarding entry found for destination node {dst_node}"
        warnings.warn(msg, stacklevel=2)
        next_node = None

    # リンク切断時にランダムにnext nodeを選択する
    # (エッジが切断されていることを知っているが, フォワーディングテーブルには反映されていない場合)
    if enable_random_routing_when_edge_failure:
        graph: Graph = node_knowledge.space_time_graph.graph_for_snapshots[time_index]
        src_node: Node = node_knowledge.target_node
        neighbor_nodes = list(graph.neighbors(src_node))
        if next_node is None or (
            {(src_node, next_node), (next_node, src_node)} & set(node_knowledge.failure_assumed_edge_list)
        ):
            next_node = rng.choice(neighbor_nodes)

    # next_nodeが過去通ってきたパスに含まれる場合に、パスに含まれない隣接ノードをランダムに選択する
    # 隣接の全てのノードが過去通ってきたパスに含まれる場合は、ランダムに選択する
    if prevent_loop and (not do_backup) and (next_node in path):
        graph = node_knowledge.space_time_graph.graph_for_snapshots[time_index]
        src_node = node_knowledge.target_node
        neighbor_nodes = list(graph.neighbors(src_node))
        if all(node in path for node in neighbor_nodes):
            next_node = rng.choice(neighbor_nodes)
        else:
            remain_neighbor_nodes: list = [node for node in neighbor_nodes if node not in path]
            next_node = rng.choice(remain_neighbor_nodes)

    return next_node



__all__ = [
    "SatToGatewayBinaryCommLinkCalculator",
    "SatToGatewayBinaryCommLinkCalculatorWithScintillation",
    "SatToGatewayBinaryMemoryCommLinkCalculator",
    "SatToGatewayStochasticBinaryCommLinkCalculator",
]

from collections.abc import Callable, Collection, Mapping, Sequence
from typing import Annotated

import numpy as np
import numpy.typing as npt
from typing_extensions import Doc

from cosmica.comm_link.uncertainty import CloudStates
from cosmica.dtos import DynamicsData
from cosmica.models import Gateway, GatewayOGS, Node, Satellite
from cosmica.utils.constants import BOLTZ_CONST, SPEED_OF_LIGHT
from cosmica.utils.coordinates import ecef2aer, geodetic2ecef
from cosmica.utils.gauss_beam import calc_gauss_rx_power
from cosmica.utils.vector import angle_between

from .base import CommLinkCalculator, CommLinkPerformance, MemorylessCommLinkCalculator
from .uncertainty import ApertureAveragedLogNormalScintillationModel, AtmosphericScintillationModel


class SatToGatewayBinaryCommLinkCalculator(MemorylessCommLinkCalculator[Satellite, Gateway]):
    """Calculate satellite-to-gateway communication link performance for each edge in a network.

    The link performance is calculated as a binary value, i.e., 1 if the link is available and 0 otherwise.
    """

    def __init__(
        self,
        *,
        satellite_to_gateway_link_capacity: float,
        sun_exclusion_angle: float = 0.0,
    ) -> None:
        self.satellite_to_gateway_link_capacity = satellite_to_gateway_link_capacity
        self.sun_exclusion_angle = sun_exclusion_angle

    def calc(
        self,
        edges: Collection[tuple[Satellite, Gateway]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,  # noqa: ARG002 For interface compatibility
    ) -> dict[tuple[Satellite, Gateway], CommLinkPerformance]:
        return {
            edge: self._calc_satellite_to_gateway(
                satellite_position_ecef=dynamics_data.satellite_position_ecef[edge[0]],
                sun_direction_ecef=dynamics_data.sun_direction_ecef,
                sun_exclusion_angle=self.sun_exclusion_angle,
                gateway=edge[1],
            )
            for edge in edges
        }

    def _calc_satellite_to_gateway(
        self,
        satellite_position_ecef: npt.NDArray[np.floating],
        gateway: Gateway,
        sun_direction_ecef: npt.NDArray[np.floating],
        sun_exclusion_angle: float = 0.0,
    ) -> CommLinkPerformance:
        assert satellite_position_ecef.shape == (3,)
        assert sun_direction_ecef.shape == (3,)
        _, elevation, srange = ecef2aer(
            x=satellite_position_ecef[0],
            y=satellite_position_ecef[1],
            z=satellite_position_ecef[2],
            lat0=gateway.latitude,
            lon0=gateway.longitude,
            h0=gateway.altitude,
            deg=False,
        )
        x0, y0, z0 = geodetic2ecef(
            lat=gateway.latitude,
            lon=gateway.longitude,
            alt=gateway.altitude,
            deg=False,
        )
        gateway_ecef = np.array([x0, y0, z0])
        gateway_to_satellite_ecef = satellite_position_ecef - gateway_ecef
        edge_sun_angle = angle_between(gateway_to_satellite_ecef, sun_direction_ecef)
        link_available = bool(elevation >= gateway.minimum_elevation and edge_sun_angle >= sun_exclusion_angle)
        return CommLinkPerformance(
            link_capacity=self.satellite_to_gateway_link_capacity if link_available else 0.0,
            delay=float(srange / SPEED_OF_LIGHT),
            link_available=link_available,
        )


class SatToGatewayBinaryMemoryCommLinkCalculator(CommLinkCalculator[Satellite, Gateway]):
    """Calculate satellite-to-gateway communication link performance for each edge in a network.

    The link performance is calculated as a binary value, i.e., 1 if the link is available and 0 otherwise.
    """

    def __init__(
        self,
        *,
        memoryless_calculator: MemorylessCommLinkCalculator[Satellite, Gateway],
        link_acquisition_time: float = 60,
        skip_link_acquisition_at_simulation_start: bool = True,
    ) -> None:
        self.memoryless_calculator = memoryless_calculator
        self.link_acquisition_time = link_acquisition_time
        self.skip_link_acquisition_at_simulation_start = skip_link_acquisition_at_simulation_start

    def calc(
        self,
        edges_time_series: Sequence[Collection[tuple[Satellite, Gateway]]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,
    ) -> list[dict[tuple[Satellite, Gateway], CommLinkPerformance]]:
        assert len(edges_time_series) == len(dynamics_data.time)

        comm_link_time_series = []

        link_acquisition_start_time: dict[tuple[Satellite, Gateway], np.datetime64] = {}
        prev_edges: frozenset[tuple[Satellite, Gateway]] = frozenset()
        for time_index, edges_snapshot in enumerate(edges_time_series):
            current_time: np.datetime64 = dynamics_data.time[time_index]

            comm_link = self.memoryless_calculator.calc(
                edges=edges_snapshot,
                dynamics_data=dynamics_data[time_index],
                rng=rng,
            )

            edges_snapshot_set = frozenset(edges_snapshot)
            new_edges = edges_snapshot_set - prev_edges
            for edge in new_edges:
                if self.skip_link_acquisition_at_simulation_start and time_index == 0:
                    link_acquisition_start_time[edge] = current_time - np.timedelta64(
                        int(self.link_acquisition_time),
                        "s",
                    )
                else:
                    link_acquisition_start_time[edge] = current_time

            disappeared_edges = prev_edges - edges_snapshot_set
            for edge in disappeared_edges:  # remove edges that disappeared
                link_acquisition_start_time.pop(edge, None)
            prev_edges = edges_snapshot_set

            for edge in edges_snapshot_set:
                if not comm_link[edge]["link_available"]:
                    link_acquisition_start_time[edge] = current_time

                within_link_acquisition: bool = False
                if edge in link_acquisition_start_time:
                    within_link_acquisition = (
                        float(
                            (current_time - link_acquisition_start_time[edge]) / np.timedelta64(1, "s"),
                        )
                        < self.link_acquisition_time
                    )
                if within_link_acquisition:
                    comm_link[edge] = CommLinkPerformance(
                        link_capacity=0.0,
                        delay=np.inf,
                        link_available=False,
                    )

            comm_link_time_series.append(comm_link)

        return comm_link_time_series


class SatToGatewayStochasticBinaryCommLinkCalculator(CommLinkCalculator[Satellite, Gateway]):
    def __init__(
        self,
        *,
        memoryless_calculator: MemorylessCommLinkCalculator[Satellite, Gateway],
        stochastic_model_factory: Callable[[Node, Node], CloudStates[np.bool_]],
    ) -> None:
        self.memoryless_calculator = memoryless_calculator
        self.stochastic_model_factory = stochastic_model_factory

    def calc(
        self,
        edges_time_series: Sequence[Collection[tuple[Satellite, Gateway]]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,
    ) -> list[dict[tuple[Satellite, Gateway], CommLinkPerformance]]:
        assert len(edges_time_series) == len(dynamics_data.time)
        all_edges = {edge for edges_snapshot in edges_time_series for edge in edges_snapshot}
        edge_interrupted = {
            edge: self.stochastic_model_factory(edge[0], edge[1]).simulate(time=dynamics_data.time, rng=rng)
            for edge in all_edges
        }

        comm_link_performance = []
        for time_index, edges_snapshot in enumerate(edges_time_series):
            comm_link = self.memoryless_calculator.calc(
                edges=edges_snapshot,
                dynamics_data=dynamics_data[time_index],
                rng=rng,
            )
            for edge in comm_link:
                if edge_interrupted[edge][time_index]:
                    comm_link[edge] = CommLinkPerformance(link_capacity=0.0, delay=0.0, link_available=False)
            comm_link_performance.append(comm_link)

        return comm_link_performance


class SatToGatewayBinaryCommLinkCalculatorWithScintillation(MemorylessCommLinkCalculator[Satellite, GatewayOGS]):
    """Calculate satellite-to-gateway communication link performance for each edge in a network with turbulence.

    The link performance is calculated as a binary value, i.e., 1 if the link is available and 0 otherwise.
    """

    def __init__(
        self,
        *,
        satellite_to_gateway_link_capacity: float,
        link_capacity: float,
        noise_figure: float,
        lna_gain: float,
        lct_p0: float,
        turbulence_model_factory: Callable[
            [GatewayOGS],
            AtmosphericScintillationModel,
        ] = lambda gateway: ApertureAveragedLogNormalScintillationModel(
            default_rytov_variance=gateway.rytov_variance,
            wavelength=1550 * 1e-9,
            aperture_diameter=gateway.aperture_size,
        ),
    ) -> None:
        self.satellite_to_gateway_link_capacity = satellite_to_gateway_link_capacity
        self.link_capacity = link_capacity
        self.noise_figure = noise_figure
        self.lna_gain = lna_gain
        self.lct_p0 = lct_p0
        self.turbulence_model_factory = turbulence_model_factory

    def calc(
        self,
        edges: Collection[tuple[Satellite, GatewayOGS]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,
    ) -> dict[tuple[Satellite, GatewayOGS], CommLinkPerformance]:
        gateway_turbulence_map = self._map_turbulence2gateway(edges)
        return {
            edge: self._calc_satellite_to_gateway(
                satellite_position_ecef=dynamics_data.satellite_position_ecef[edge[0]],
                sun_direction_ecef=dynamics_data.sun_direction_ecef,
                sun_exclusion_angle=0.0,
                rng=rng,
                gateway=edge[1],
                gateway_turbulence_map=gateway_turbulence_map,
            )
            for edge in edges
        }

    def _map_turbulence2gateway(
        self,
        edges: Collection[tuple[Satellite, GatewayOGS]],
    ) -> dict[GatewayOGS, AtmosphericScintillationModel]:
        gateway_turbulence_map: dict[GatewayOGS, AtmosphericScintillationModel] = {}
        for edge in edges:
            if gateway_turbulence_map.get(edge[1]) is None:
                gateway_turbulence_map[edge[1]] = self.turbulence_model_factory(edge[1])
        return gateway_turbulence_map

    def _calc_capacity(
        self,
        distance: Annotated[
            float,
            Doc("Distance between satellite and OGS"),
        ],
        gateway: Annotated[
            GatewayOGS,
            Doc("Gateway instance in the satellite-gateway link"),
        ],
        turbulence_map: Annotated[
            Mapping[GatewayOGS, AtmosphericScintillationModel],
            Doc("Dictionary mapping each OGS to its respective turbulence model."),
        ],
        rng: np.random.Generator,
        rytov_variance: Annotated[
            float | None,
            Doc("Rytov variance value"),
        ] = None,
    ) -> float:
        scintillation_factor = turbulence_map[gateway].sample(
            rng=rng,
            link_distance=distance,
            rytov_variance=rytov_variance,
        )
        # Downlink Capacity Calculation
        power = calc_gauss_rx_power(self.lct_p0, gateway.aperture_size, distance)
        t_sys = 300  # K
        noise_factor = 10 ** (self.noise_figure / 10)
        noise = t_sys * self.link_capacity * BOLTZ_CONST * noise_factor
        gain = 10 ** (self.lna_gain / 10)  # convert from dB to W
        snr = gain * power * scintillation_factor / (noise)
        return self.link_capacity * np.log2(1 + snr)

    def _calc_satellite_to_gateway(
        self,
        satellite_position_ecef: npt.NDArray[np.floating],
        gateway: GatewayOGS,
        gateway_turbulence_map: Mapping[GatewayOGS, AtmosphericScintillationModel],
        sun_direction_ecef: npt.NDArray[np.floating],
        rng: np.random.Generator,
        sun_exclusion_angle: float = 0.0,
    ) -> CommLinkPerformance:
        assert satellite_position_ecef.shape == (3,)
        assert sun_direction_ecef.shape == (3,)
        _, elevation, srange = ecef2aer(
            x=satellite_position_ecef[0],
            y=satellite_position_ecef[1],
            z=satellite_position_ecef[2],
            lat0=gateway.latitude,
            lon0=gateway.longitude,
            h0=gateway.altitude,
            deg=False,
        )
        x0, y0, z0 = geodetic2ecef(
            lat=gateway.latitude,
            lon=gateway.longitude,
            alt=gateway.altitude,
            deg=False,
        )
        gateway_ecef = np.array([x0, y0, z0])
        gateway_to_satellite_ecef = satellite_position_ecef - gateway_ecef
        distance = np.linalg.norm(gateway_to_satellite_ecef)
        edge_sun_angle = angle_between(gateway_to_satellite_ecef, sun_direction_ecef)
        link_available = bool(elevation >= gateway.minimum_elevation and edge_sun_angle >= sun_exclusion_angle)
        return CommLinkPerformance(
            link_capacity=self._calc_capacity(float(distance), gateway, gateway_turbulence_map, rng)
            if link_available
            else 0.0,
            delay=float(srange / SPEED_OF_LIGHT),
            link_available=link_available,
        )



__all__ = [
    "calc_dcm_eci2ecef",
    "ecef2aer",
    "geodetic2ecef",
    "greenwichsrt",
    "juliandate",
]
from typing import overload

import numpy as np
import numpy.typing as npt
import pandas as pd
from pymap3d.aer import ecef2aer  # re-exported as cosmica.utils.coordinates.ecef2aer
from pymap3d.ecef import geodetic2ecef
from pymap3d.sidereal import greenwichsrt as _greenwichsrt


@overload
def juliandate(time: np.datetime64) -> float: ...


@overload
def juliandate(time: npt.NDArray[np.datetime64]) -> npt.NDArray[np.float64]: ...


def juliandate(time):
    return pd.to_datetime(time).to_julian_date()


@overload
def greenwichsrt(time: np.datetime64) -> float: ...


@overload
def greenwichsrt(time: npt.NDArray[np.datetime64]) -> npt.NDArray[np.float64]: ...


def greenwichsrt(time):
    return _greenwichsrt(juliandate(time))


def calc_dcm_eci2ecef(time: np.datetime64 | npt.NDArray[np.datetime64]) -> npt.NDArray[np.float64]:
    """Calculate the direction cosine matrix from ECI to ECEF.

    Input is UTC time.
    If the input is a single time, the output is a 3x3 matrix.
    If the input is an array of times, the output is a Nx3x3 array.
    """
    gst = greenwichsrt(time)
    if isinstance(gst, float):
        return np.array(
            [
                [np.cos(gst), np.sin(gst), 0],
                [-np.sin(gst), np.cos(gst), 0],
                [0, 0, 1],
            ],
        )
    else:
        return np.stack(
            [
                np.stack(
                    [
                        np.cos(gst),
                        np.sin(gst),
                        np.zeros_like(gst),
                    ],
                    axis=-1,
                ),
                np.stack(
                    [
                        -np.sin(gst),
                        np.cos(gst),
                        np.zeros_like(gst),
                    ],
                    axis=-1,
                ),
                np.stack(
                    [
                        np.zeros_like(gst),
                        np.zeros_like(gst),
                        np.ones_like(gst),
                    ],
                    axis=-1,
                ),
            ],
            axis=-2,
        )



__all__ = [
    "ForwardingTable",
    "ForwardingTableInformation",
]

from dataclasses import dataclass, field

from cosmica.models.node import Node


@dataclass(kw_only=True, slots=True)
class ForwardingTableInformation:
    destination: Node
    next_node: Node
    cost: float = 0.0


@dataclass(kw_only=True, slots=True)
class ForwardingTable:
    entries: dict[Node, ForwardingTableInformation] = field(default_factory=dict)

    def update_entry(
        self,
        destination: Node,
        next_node: Node,
        cost: float = 0.0,
    ) -> None:
        if destination in self.entries:
            entry: ForwardingTableInformation = self.entries[destination]
            entry.next_node = next_node
            entry.cost = cost
        else:
            self.entries[destination] = ForwardingTableInformation(
                destination=destination,
                next_node=next_node,
                cost=cost,
            )

    def remove_entry(
        self,
        destination: Node,
    ) -> None:
        if destination in self.entries:
            del self.entries[destination]

    def find_entry(
        self,
        destination: Node,
    ) -> ForwardingTableInformation | None:
        return self.entries.get(destination, None)



__all__ = [
    "ConstellationSatellite",
    "Satellite",
    "SatelliteTerminal",
    "UserSatellite",
]
from abc import ABC
from collections.abc import Hashable
from dataclasses import dataclass
from typing import override

from .node import Node
from .orbit import SatelliteOrbitModel
from .terminal import OpticalCommunicationTerminal


class Satellite[T: Hashable](Node[T], ABC): ...


@dataclass(frozen=True, kw_only=True, slots=True)
class ConstellationSatellite[T: Hashable](Satellite[T]):
    id: T

    # Allow none for backward compatibility.
    orbit: SatelliteOrbitModel | None = None

    @classmethod
    @override
    def class_name(cls) -> str:
        return "CSAT"


@dataclass(frozen=True, kw_only=True, slots=True)
class UserSatellite[T: Hashable](Satellite[T]):
    id: T

    @classmethod
    @override
    def class_name(cls) -> str:
        return "USAT"


@dataclass(frozen=True, kw_only=True, slots=True)
class SatelliteTerminal[T: Hashable](Satellite[T]):
    id: T
    terminal_id: T
    azimuth_min: float
    azimuth_max: float
    elevation_min: float
    elevation_max: float
    angular_velocity_max: float

    @property
    def terminal(self) -> OpticalCommunicationTerminal[T]:
        return OpticalCommunicationTerminal(
            id=self.terminal_id,
            azimuth_min=self.azimuth_min,
            azimuth_max=self.azimuth_max,
            elevation_min=self.elevation_min,
            elevation_max=self.elevation_max,
            angular_velocity_max=self.angular_velocity_max,
        )



from __future__ import annotations

__all__ = [
    "Gateway",
    "GatewayOGS",
]
import tomllib
from collections.abc import Hashable
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Self, override

from typing_extensions import deprecated

if TYPE_CHECKING:
    from collections.abc import MutableMapping


import numpy as np

from .node import Node


@dataclass(frozen=True, kw_only=True, slots=True)
class Gateway[T: Hashable](Node[T]):
    id: T
    latitude: float = field(compare=False)
    longitude: float = field(compare=False)
    minimum_elevation: float = field(compare=False)
    altitude: float = field(default=0.0, compare=False)
    n_terminals: int = field(default=1, compare=False)

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def from_toml_file(cls, toml_file_path: Path | str) -> list[Self]:
        toml_file_path = Path(toml_file_path)
        with toml_file_path.open("rb") as f:
            toml_data = tomllib.load(f)
        return list(map(cls.parse_gateway_item, toml_data["gateways"]))

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_gateway_item(cls, item: MutableMapping[str, Any]) -> Self:
        item["latitude"] = np.deg2rad(item.pop("lat_deg"))
        item["longitude"] = np.deg2rad(item.pop("lon_deg"))
        item["minimum_elevation"] = np.deg2rad(item.pop("min_el_deg"))
        return cls(**item)

    @classmethod
    @override
    def class_name(cls) -> str:
        return "GW"


@dataclass(frozen=True, kw_only=True, slots=True)
class GatewayOGS[T: Hashable](Node[T]):
    id: T
    latitude: float = field(compare=False)
    longitude: float = field(compare=False)
    minimum_elevation: float = field(compare=False)
    altitude: float = field(default=0.0, compare=False)
    n_terminals: int = field(default=1, compare=False)
    aperture_size: float = field(default=1.0, compare=False)
    rytov_variance: float = field(default=0.5, compare=False)

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def from_toml_file(cls, toml_file_path: Path | str) -> list[Self]:
        toml_file_path = Path(toml_file_path)
        with toml_file_path.open("rb") as f:
            toml_data = tomllib.load(f)
        return list(map(cls.parse_gateway_item, toml_data["gateways"]))

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_gateway_item(cls, item: MutableMapping[str, Any]) -> Self:
        item["latitude"] = np.deg2rad(item.pop("lat_deg"))
        item["longitude"] = np.deg2rad(item.pop("lon_deg"))
        item["minimum_elevation"] = np.deg2rad(item.pop("min_el_deg"))
        item["aperture_size"] = item.pop("aperture_size_m")
        item["rytov_variance"] = item.pop("rytov_variance")

        return cls(**item)

    @classmethod
    @override
    def class_name(cls) -> str:
        return "GW_OGS"



__all__ = [
    "OTC2OTCBinaryCommLinkCalculator",
    "SatToSatBinaryCommLinkCalculator",
    "SatToSatBinaryMemoryCommLinkCalculator",
]

import logging
from collections.abc import Collection, Sequence
from itertools import chain
from typing import Annotated

import numpy as np
import numpy.typing as npt
from numpy.random import Generator
from typing_extensions import Doc

from cosmica.dtos import DynamicsData
from cosmica.models import OpticalCommunicationTerminal, Satellite, SatelliteTerminal
from cosmica.utils.constants import SPEED_OF_LIGHT
from cosmica.utils.vector import angle_between, is_satellite_in_eclipse, normalize, unit_vector_to_azimuth_elevation

from .base import CommLinkCalculator, CommLinkPerformance, MemorylessCommLinkCalculator

logger = logging.getLogger(__name__)


class SatToSatBinaryCommLinkCalculator(MemorylessCommLinkCalculator[Satellite, Satellite]):
    """Calculate satellite-to-satellite communication link performance for each edge in a network.

    The link performance is calculated as a binary value, i.e., 1 if the link is available and 0 otherwise.
    """

    def __init__(
        self,
        *,
        inter_satellite_link_capacity: float,
        max_inter_satellite_distance: float = float("inf"),
        lowest_altitude: float = 0.0,
        max_relative_angular_velocity: float = float("inf"),
        sun_exclusion_angle: float = 0.0,
    ) -> None:
        self.max_inter_satellite_distance = max_inter_satellite_distance
        self.inter_satellite_link_capacity = inter_satellite_link_capacity
        self.lowest_altitude = lowest_altitude
        self.max_relative_angular_velocity = max_relative_angular_velocity
        self.sun_exclusion_angle = sun_exclusion_angle

    def calc(
        self,
        edges: Collection[tuple[Satellite, Satellite]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,  # noqa: ARG002 For interface compatibility
    ) -> dict[tuple[Satellite, Satellite], CommLinkPerformance]:
        return {
            edge: self._calc_satellite_to_satellite(
                positions_eci=(
                    dynamics_data.satellite_position_eci[edge[0]],
                    dynamics_data.satellite_position_eci[edge[1]],
                ),
                velocities_eci=(
                    dynamics_data.satellite_velocity_eci[edge[0]],
                    dynamics_data.satellite_velocity_eci[edge[1]],
                ),
                attitude_angular_velocities_eci=(
                    dynamics_data.satellite_attitude_angular_velocity_eci[edge[0]],
                    dynamics_data.satellite_attitude_angular_velocity_eci[edge[1]],
                ),
                sun_direction_eci=dynamics_data.sun_direction_eci,
            )
            for edge in edges
        }

    def _calc_satellite_to_satellite(
        self,
        *,
        positions_eci: Annotated[
            tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
            Doc("Position vectors in ECI frame. Shape: (3,)"),
        ],
        velocities_eci: Annotated[
            tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
            Doc("Velocity vectors in ECI frame. Shape: (3,)"),
        ],
        attitude_angular_velocities_eci: tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
        sun_direction_eci: Annotated[
            npt.NDArray[np.floating],
            Doc("Sun direction vector in ECI frame. Shape: (3,)"),
        ],
    ) -> CommLinkPerformance:
        """Calculate binary communication link performance between two satellites."""
        for vec in chain(positions_eci, velocities_eci):
            assert vec.shape == (3,), f"Position and velocity vectors must be 3-dimensional, but got shape {vec.shape}"

        distance = float(np.linalg.norm(positions_eci[1] - positions_eci[0]))

        # Calculate relative angular velocity due to translational motion
        # The calculated angular velocity is that of the second satellite relative to the first satellite.
        relative_position_eci = positions_eci[1] - positions_eci[0]
        relative_velocity_eci = velocities_eci[1] - velocities_eci[0]
        relative_angular_velocity_translational_eci = (
            np.cross(relative_position_eci, relative_velocity_eci) / distance**2
        )

        relative_angular_velocities = (
            relative_angular_velocity_translational_eci - attitude_angular_velocities_eci[0],
            # Note: The relative angular velocity should be negated to get the angular velocity of the first satellite
            # relative to the second satellite.
            -relative_angular_velocity_translational_eci - attitude_angular_velocities_eci[1],
        )

        # Check if either satellite is in eclipse - if so, ignore sun exclusion angle for that satellite
        satellite_a_in_eclipse = is_satellite_in_eclipse(positions_eci[0], sun_direction_eci)
        satellite_b_in_eclipse = is_satellite_in_eclipse(positions_eci[1], sun_direction_eci)

        # Calculate sun exclusion angle constraints for each direction
        # If satellite is in eclipse, skip sun exclusion angle check for that direction
        sun_exclusion_satisfied = True

        if not satellite_a_in_eclipse:
            # Check sun exclusion angle from satellite A's perspective (A looking towards B)
            edge_sun_angle_a_to_b = angle_between(relative_position_eci, sun_direction_eci)
            if edge_sun_angle_a_to_b < self.sun_exclusion_angle:
                sun_exclusion_satisfied = False

        if not satellite_b_in_eclipse:
            # Check sun exclusion angle from satellite B's perspective (B looking towards A)
            edge_sun_angle_b_to_a = angle_between(-relative_position_eci, sun_direction_eci)
            if edge_sun_angle_b_to_a < self.sun_exclusion_angle:
                sun_exclusion_satisfied = False

        link_available = bool(
            distance < self.max_inter_satellite_distance
            and all(
                float(np.linalg.norm(relative_angular_velocity)) < self.max_relative_angular_velocity
                for relative_angular_velocity in relative_angular_velocities
            )
            and sun_exclusion_satisfied,
        )

        return CommLinkPerformance(
            link_capacity=self.inter_satellite_link_capacity if link_available else 0.0,
            delay=float(distance / SPEED_OF_LIGHT),
            link_available=link_available,
        )


class SatToSatBinaryMemoryCommLinkCalculator(CommLinkCalculator[Satellite, Satellite]):
    """Calculate satellite-to-satellite communication link performance for each edge in a network.

    The link performance is calculated as a binary value, i.e., 1 if the link is available and 0 otherwise.
    """

    def __init__(
        self,
        *,
        memoryless_calculator: MemorylessCommLinkCalculator[Satellite, Satellite],
        link_acquisition_time: float = 60.0,
        skip_link_acquisition_at_simulation_start: bool = True,
    ) -> None:
        self.memoryless_calculator = memoryless_calculator
        self.link_acquisition_time = link_acquisition_time
        self.skip_link_acquisition_at_simulation_start = skip_link_acquisition_at_simulation_start

    def calc(  # noqa: C901
        self,
        edges_time_series: Sequence[Collection[tuple[Satellite, Satellite]]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,
    ) -> list[dict[tuple[Satellite, Satellite], CommLinkPerformance]]:
        assert len(edges_time_series) == len(dynamics_data.time)

        comm_link_time_series: list[dict[tuple[Satellite, Satellite], CommLinkPerformance]] = []

        # ― per-edge state ―
        # TODO(): Currently only undirected edges are targeted, so inter-satellite links are represented in frozenset
        # prev_dot: dict[frozenset[Satellite], float] = {}
        link_acquisition_start_time: dict[frozenset[Satellite], np.datetime64] = {}
        prev_edges: frozenset[frozenset[Satellite]] = frozenset()

        for time_index, edges_snapshot in enumerate(edges_time_series):
            current_time: np.datetime64 = dynamics_data.time[time_index]

            comm_link = self.memoryless_calculator.calc(
                edges=edges_snapshot,
                dynamics_data=dynamics_data[time_index],
                rng=rng,
            )

            edges_snapshot_set = to_edge_frozenset(edges_snapshot)

            # ── update “first-seen” bookkeeping ──────────────────── ★
            new_edges = edges_snapshot_set - prev_edges
            for edge_fs in new_edges:
                if self.skip_link_acquisition_at_simulation_start and time_index == 0:
                    link_acquisition_start_time[edge_fs] = current_time - np.timedelta64(
                        int(self.link_acquisition_time),
                        "s",
                    )
                else:
                    link_acquisition_start_time[edge_fs] = current_time

            disappeared_edges = prev_edges - edges_snapshot_set
            for edge_fs in disappeared_edges:  # フェードアウトしたら状態を消去
                link_acquisition_start_time.pop(edge_fs, None)
                # prev_dot.pop(edge_fs, None)
            prev_edges = edges_snapshot_set
            # ───────────────────────────────────────────────────────

            for edge_fs in edges_snapshot_set:
                sat_a, sat_b = sorted(edge_fs, key=id)

                if ((sat_a, sat_b) in comm_link and comm_link[(sat_a, sat_b)]["link_available"] is False) or (
                    (sat_b, sat_a) in comm_link and comm_link[(sat_b, sat_a)]["link_available"] is False
                ):
                    link_acquisition_start_time[edge_fs] = current_time

                # --- link acquisition delay ------------ ★
                within_link_acquisition: bool = False
                if edge_fs in link_acquisition_start_time:
                    within_link_acquisition = (
                        float(
                            (current_time - link_acquisition_start_time[edge_fs]) / np.timedelta64(1, "s"),
                        )
                        < self.link_acquisition_time
                    )

                if within_link_acquisition:
                    if (sat_a, sat_b) in comm_link:
                        comm_link[(sat_a, sat_b)] = CommLinkPerformance(
                            link_capacity=0.0,
                            delay=np.inf,
                            link_available=False,
                        )
                    elif (sat_b, sat_a) in comm_link:
                        comm_link[(sat_b, sat_a)] = CommLinkPerformance(
                            link_capacity=0.0,
                            delay=np.inf,
                            link_available=False,
                        )
                # ----------------------------------------------------

            comm_link_time_series.append(comm_link)

        return comm_link_time_series


class OTC2OTCBinaryCommLinkCalculator(CommLinkCalculator[SatelliteTerminal, SatelliteTerminal]):
    """Calculate satellite-to-satellite communication link performance for each terminal in a network.

    The link performance is calculated as a binary value, i.e., 1 if the link is available and 0 otherwise.
    """

    def __init__(
        self,
        *,
        inter_satellite_link_capacity: float,
        max_inter_satellite_distance: float = float("inf"),
        lowest_altitude: float = 0.0,
        max_relative_angular_velocity: float = float("inf"),
        sun_exclusion_angle: float = 0.0,
    ) -> None:
        self.max_inter_satellite_distance = max_inter_satellite_distance
        self.inter_satellite_link_capacity = inter_satellite_link_capacity
        self.lowest_altitude = lowest_altitude
        self.max_relative_angular_velocity = max_relative_angular_velocity
        self.sun_exclusion_angle = sun_exclusion_angle

    def calc(
        self,
        edges_time_series: Sequence[Collection[tuple[SatelliteTerminal, SatelliteTerminal]]],
        *,
        dynamics_data: DynamicsData,
        rng: Generator,  # noqa: ARG002
    ) -> list[dict[tuple[SatelliteTerminal, SatelliteTerminal], CommLinkPerformance]]:
        terminal_memo: dict[tuple[SatelliteTerminal, SatelliteTerminal], list[tuple[float, float]]] = {}
        comm_link_time_series = []
        prev_time = dynamics_data.time[0]

        for i, snapshot in enumerate(edges_time_series):
            edges_performance = {}
            current_time = dynamics_data.time[i]
            time_delta = current_time - prev_time if i != 0 else 1
            if time_delta == 0:
                msg = "time_delta must be non-zero"
                raise ValueError(msg)
            # Note: the terminal angular velocity and pointing verifications in the first iteration will return
            # meaningless results and should be disconsidered during analysis.

            for edge in snapshot:
                # TODO(): エッジがtupleとして定義されているので、順序を考慮して調べる必要がある
                previous_terminal_directions = terminal_memo.get(edge, [(0.0, 0.0), (0.0, 0.0)])

                # TODO(): おそらく dynamics_data の中で各タイムステップの値を取ってくる必要がある
                comm_link_performance, terminal_directions = self._calc_satellite_to_satellite(
                    positions_eci=(
                        dynamics_data.satellite_position_eci[edge[0]],
                        dynamics_data.satellite_position_eci[edge[1]],
                    ),
                    velocities_eci=(
                        dynamics_data.satellite_velocity_eci[edge[0]],
                        dynamics_data.satellite_velocity_eci[edge[1]],
                    ),
                    attitude_angular_velocities_eci=(
                        dynamics_data.satellite_attitude_angular_velocity_eci[edge[0]],
                        dynamics_data.satellite_attitude_angular_velocity_eci[edge[1]],
                    ),
                    sun_direction_eci=dynamics_data.sun_direction_eci,
                    terminals=(
                        edge[0].terminal,
                        edge[1].terminal,
                    ),
                    previous_terminal_directions=previous_terminal_directions,
                    time_delta=time_delta,
                )

                edges_performance[edge] = comm_link_performance
                terminal_memo[edge] = terminal_directions

            comm_link_time_series.append(edges_performance)
            prev_time = current_time

        return comm_link_time_series

    def _calc_satellite_to_satellite(
        self,
        *,
        positions_eci: Annotated[
            tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
            Doc("Position vectors in ECI frame. Shape: (3,)"),
        ],
        velocities_eci: Annotated[
            tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
            Doc("Velocity vectors in ECI frame. Shape: (3,)"),
        ],
        attitude_angular_velocities_eci: tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
        sun_direction_eci: Annotated[
            npt.NDArray[np.floating],
            Doc("Sun direction vector in ECI frame. Shape: (3,)"),
        ],
        terminals: Annotated[
            tuple[OpticalCommunicationTerminal, OpticalCommunicationTerminal],
            Doc("Optical Communication Terminals of pair of satellites"),
        ],
        previous_terminal_directions: Annotated[
            list[tuple[float, float]],
            Doc("Azimuth and elevation values for each terminal in the previous time step"),
        ],
        time_delta: Annotated[
            float,
            Doc("Float representing time difference to previous configuration in the time series"),
        ],
    ) -> tuple[CommLinkPerformance, list[tuple[float, float]]]:
        """Calculate binary communication link performance between two satellites."""
        for vec in chain(positions_eci, velocities_eci):
            assert vec.shape == (3,), f"Position and velocity vectors must be 3-dimensional, but got shape {vec.shape}"

        distance = float(np.linalg.norm(positions_eci[1] - positions_eci[0]))

        # Calculate relative angular velocity due to translational motion
        # The calculated angular velocity is that of the second satellite relative to the first satellite.
        relative_position_eci = positions_eci[1] - positions_eci[0]
        relative_velocity_eci = velocities_eci[1] - velocities_eci[0]
        relative_angular_velocity_translational_eci = (
            np.cross(relative_position_eci, relative_velocity_eci) / distance**2
        )

        relative_angular_velocities = (
            relative_angular_velocity_translational_eci - attitude_angular_velocities_eci[0],
            # Note: The relative angular velocity should be negated to get the angular velocity of the first satellite
            # relative to the second satellite.
            -relative_angular_velocity_translational_eci - attitude_angular_velocities_eci[1],
        )

        # Check if either satellite is in eclipse - if so, ignore sun exclusion angle for that satellite
        satellite_a_in_eclipse = is_satellite_in_eclipse(positions_eci[0], sun_direction_eci)
        satellite_b_in_eclipse = is_satellite_in_eclipse(positions_eci[1], sun_direction_eci)

        # Calculate sun exclusion angle constraints for each direction
        # If satellite is in eclipse, skip sun exclusion angle check for that direction
        sun_exclusion_satisfied = True

        if not satellite_a_in_eclipse:
            # Check sun exclusion angle from satellite A's perspective (A looking towards B)
            edge_sun_angle_a_to_b = angle_between(relative_position_eci, sun_direction_eci)
            if edge_sun_angle_a_to_b < self.sun_exclusion_angle:
                sun_exclusion_satisfied = False

        if not satellite_b_in_eclipse:
            # Check sun exclusion angle from satellite B's perspective (B looking towards A)
            edge_sun_angle_b_to_a = angle_between(-relative_position_eci, sun_direction_eci)
            if edge_sun_angle_b_to_a < self.sun_exclusion_angle:
                sun_exclusion_satisfied = False

        terminal_directions = self._calc_terminal_directions(relative_position_eci)
        terminal_angular_velocity = [
            [(terminal[0] - previous_terminal[0]) / time_delta, (terminal[1] - previous_terminal[1]) / time_delta]
            for terminal, previous_terminal in zip(terminal_directions, previous_terminal_directions, strict=False)
        ]
        link_available = bool(
            distance < self.max_inter_satellite_distance
            and all(
                float(np.linalg.norm(relative_angular_velocity)) < self.max_relative_angular_velocity
                for relative_angular_velocity in relative_angular_velocities
            )
            and sun_exclusion_satisfied
            and all(
                # For terminal direction checks, a more careful implementation is required
                # using the appropriate coordinate frame transformations
                # terminal_directions[i][0] < terminal.azimuth_max
                # and terminal_directions[i][0] > terminal.azimuth_min
                # and terminal_directions[i][1] > terminal.elevation_min
                # and terminal_directions[i][1] < terminal.elevation_max
                terminal_angular_velocity[i][0] < terminal.angular_velocity_max
                and terminal_angular_velocity[i][1] < terminal.angular_velocity_max
                for i, terminal in enumerate(terminals)
            ),
        )

        return (
            CommLinkPerformance(
                link_capacity=self.inter_satellite_link_capacity if link_available else 0.0,
                delay=float(distance / SPEED_OF_LIGHT),
                link_available=link_available,
            ),
            terminal_directions,
        )

    def _calc_terminal_directions(self, relative_position: npt.NDArray) -> list[tuple[float, float]]:
        unitary_direction = normalize(relative_position)
        terminal_a = unit_vector_to_azimuth_elevation(unitary_direction)
        terminal_b = unit_vector_to_azimuth_elevation(
            np.concatenate((unitary_direction[:-1], [-unitary_direction[-1]])),
        )
        # Note: Here we only inverted the z component of the unit vector, as it is assumed the terminals are located
        # at opposite faces
        return [terminal_a, terminal_b]


def to_edge_frozenset(snapshot: Collection[tuple[Satellite, Satellite]]) -> frozenset[frozenset[Satellite]]:
    """Convert a snapshot to {frozenset{SatA, SatB}, …} representation."""
    return frozenset(frozenset(edge) for edge in snapshot)



from __future__ import annotations

__all__ = [
    "CircularSatelliteOrbit",
    "CircularSatelliteOrbitPropagator",
    "EllipticalSatelliteOrbit",
    "EllipticalSatelliteOrbitPropagator",
    "SatelliteOrbit",
    "SatelliteOrbitState",
    "make_satellite_orbit",
]

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Literal, override

import numpy as np
from sgp4.model import Satrec
from skyfield.api import EarthSatellite, Time, Timescale, load, utc
from typing_extensions import deprecated

from cosmica.models import CircularSatelliteOrbitModel, EllipticalSatelliteOrbitModel, GravityModel
from cosmica.utils.constants import EARTH_MU
from cosmica.utils.vector import rowwise_matmul

if TYPE_CHECKING:
    import numpy.typing as npt


logger = logging.getLogger(__name__)

ReferenceFrame = Literal["teme", "j2000", "gcrs"]


@dataclass(frozen=True, kw_only=True, slots=True)
class SatelliteOrbitState:
    position_eci: npt.NDArray[np.floating]
    velocity_eci: npt.NDArray[np.floating]

    def __post_init__(self) -> None:
        assert self.position_eci.shape[-1] == 3, self.position_eci.shape
        assert self.velocity_eci.shape[-1] == 3, self.velocity_eci.shape
        assert self.position_eci.shape[:-1] == self.velocity_eci.shape[:-1], (
            self.position_eci.shape,
            self.velocity_eci.shape,
        )

    def __getitem__(self, item: int | slice) -> SatelliteOrbitState:
        return SatelliteOrbitState(
            position_eci=self.position_eci[item],
            velocity_eci=self.velocity_eci[item],
        )

    def calc_position_ecef(self, dcm_eci_to_ecef: npt.NDArray[np.floating]) -> npt.NDArray[np.floating]:
        return np.einsum("ijk,ik->ij", dcm_eci_to_ecef, self.position_eci)


class SatelliteOrbitPropagator(ABC):
    @abstractmethod
    def propagate(self, time: npt.NDArray[np.datetime64]) -> SatelliteOrbitState: ...


class CircularSatelliteOrbitPropagator(SatelliteOrbitPropagator):
    def __init__(self, model: CircularSatelliteOrbitModel) -> None:
        self._model = model

    @cached_property
    def mean_motion(self) -> float:
        return float(np.sqrt(EARTH_MU / self._model.semi_major_axis**3))  # pyright: ignore [reportAny]

    @cached_property
    def dcm_orbit_to_eci(self) -> npt.NDArray[np.floating]:
        return np.array(
            [
                [np.cos(self._model.raan), -np.sin(self._model.raan), 0],
                [np.sin(self._model.raan), np.cos(self._model.raan), 0],
                [0, 0, 1],
            ],
            dtype=np.float64,
        ) @ np.array(
            [
                [1, 0, 0],
                [0, np.cos(self._model.inclination), -np.sin(self._model.inclination)],
                [0, np.sin(self._model.inclination), np.cos(self._model.inclination)],
            ],
            dtype=np.float64,
        )

    def propagate(self, time: npt.NDArray[np.datetime64]) -> SatelliteOrbitState:
        time_from_epoch = time - self._model.epoch
        return self.propagate_from_epoch(time_from_epoch)

    def propagate_from_epoch(self, time_from_epoch: npt.NDArray[np.timedelta64]) -> SatelliteOrbitState:
        time_from_epoch_in_seconds = time_from_epoch / np.timedelta64(1, "s")
        phase = self._model.phase_at_epoch + self.mean_motion * time_from_epoch_in_seconds
        in_plane_position = self._model.semi_major_axis * np.array([np.cos(phase), np.sin(phase), np.zeros_like(phase)])
        in_plane_velocity = (
            self.mean_motion
            * self._model.semi_major_axis
            * np.array([-np.sin(phase), np.cos(phase), np.zeros_like(phase)])
        )

        position_eci: npt.NDArray[np.floating] = rowwise_matmul(self.dcm_orbit_to_eci, in_plane_position.T)
        velocity_eci: npt.NDArray[np.floating] = rowwise_matmul(self.dcm_orbit_to_eci, in_plane_velocity.T)

        return SatelliteOrbitState(position_eci=position_eci, velocity_eci=velocity_eci)


class EllipticalSatelliteOrbitPropagator(SatelliteOrbitPropagator):
    def __init__(self, model: EllipticalSatelliteOrbitModel, reference_frame: ReferenceFrame = "gcrs") -> None:
        self._model = model
        rf = reference_frame.lower()
        if rf not in {"teme", "j2000", "gcrs"}:
            logger.error("Invalid reference frame: %s", reference_frame)
            msg = f"Invalid reference_frame: {reference_frame!r}"
            raise ValueError(msg)
        self.reference_frame = rf

    @cached_property
    def ts(self) -> Timescale:
        return load.timescale()

    @cached_property
    def mean_motion(self) -> float:
        return np.sqrt(EARTH_MU / self._model.semi_major_axis**3)

    def datetime64_utc_to_skytime(self, t_datetime64: npt.NDArray[np.datetime64]) -> Time:
        t_datetimes = [time.astype(datetime).replace(tzinfo=utc) for time in t_datetime64]
        return self.ts.from_datetimes(t_datetimes)

    @cached_property
    def satellite(self) -> EarthSatellite:
        satrec = Satrec()
        sgp4_ref_epoch = np.datetime64("1949-12-31T00:00:00")
        epoch = (self._model.epoch - sgp4_ref_epoch) / np.timedelta64(1, "D")
        rad_per_min = 60
        satrec.sgp4init(
            self._model.gravity_model.value,
            "i",
            self._model.satnum,
            epoch,
            self._model.drag_coeff,
            0.0,
            0.0,
            self._model.eccentricity,
            self._model.argpo,
            self._model.inclination,
            self._model.phase_at_epoch,
            self.mean_motion * rad_per_min,
            self._model.raan,
        )
        return EarthSatellite.from_satrec(satrec, self.ts)

    @override
    def propagate(self, time: npt.NDArray[np.datetime64]) -> SatelliteOrbitState:
        time_from_epoch = self.datetime64_utc_to_skytime(time)
        if self.reference_frame in {"j2000", "gcrs"}:
            geoc = self.satellite.at(time_from_epoch)  # Geocentric position in GCRS/ICRF axes
            position_eci_km = geoc.position.km
            velocity_eci_km_s = geoc.velocity.km_per_s
            return SatelliteOrbitState(position_eci=position_eci_km.T * 1e3, velocity_eci=velocity_eci_km_s.T * 1e3)
        else:  # TEME
            [position_eci, velocity_eci, _] = self.satellite._position_and_velocity_TEME_km(time_from_epoch)  # noqa: SLF001
            return SatelliteOrbitState(position_eci=position_eci.T * 1e3, velocity_eci=velocity_eci.T * 1e3)


@deprecated("Use SatelliteOrbitModel and SatelliteOrbitPropagator instead.")
class SatelliteOrbit(ABC):
    @abstractmethod
    def propagate(self, time: npt.NDArray[np.datetime64]) -> SatelliteOrbitState: ...


@deprecated("Use CircularSatelliteOrbitModel and CircularSatelliteOrbitPropagator instead.")
@dataclass(frozen=True, kw_only=True, slots=True)
class CircularSatelliteOrbit(SatelliteOrbit):
    semi_major_axis: float
    inclination: float
    raan: float
    phase_at_epoch: float
    epoch: np.datetime64

    _model: CircularSatelliteOrbitModel = field(init=False, repr=False)
    _propagator: CircularSatelliteOrbitPropagator = field(init=False, repr=False)

    def __post_init__(self) -> None:
        object.__setattr__(
            self,
            "_model",
            CircularSatelliteOrbitModel(
                semi_major_axis=self.semi_major_axis,
                inclination=self.inclination,
                raan=self.raan,
                phase_at_epoch=self.phase_at_epoch,
                epoch=self.epoch,
            ),
        )
        object.__setattr__(self, "_propagator", CircularSatelliteOrbitPropagator(self._model))

    @cached_property
    def mean_motion(self) -> float:
        return self._propagator.mean_motion

    @cached_property
    def dcm_orbit_to_eci(self) -> npt.NDArray[np.floating]:
        return self._propagator.dcm_orbit_to_eci

    def propagate(self, time: npt.NDArray[np.datetime64]) -> SatelliteOrbitState:
        return self._propagator.propagate(time)

    def propagate_from_epoch(self, time_from_epoch: npt.NDArray[np.timedelta64]) -> SatelliteOrbitState:
        return self._propagator.propagate_from_epoch(time_from_epoch)


@deprecated(
    "Use EllipticalSatelliteOrbitModel and EllipticalSatelliteOrbitPropagator instead.",
)
@dataclass(frozen=True, kw_only=True, slots=True)
class EllipticalSatelliteOrbit(SatelliteOrbit):
    semi_major_axis: float
    inclination: float
    raan: float
    phase_at_epoch: float
    epoch: np.datetime64
    satnum: int  # Satellite number
    gravity_model: GravityModel = GravityModel.WGS84
    drag_coeff: float  # B star drag coefficient given in [1/earth radii]
    eccentricity: float
    argpo: float  # Argument of Perigee

    _model: EllipticalSatelliteOrbitModel = field(init=False, repr=False)
    _propagator: EllipticalSatelliteOrbitPropagator = field(init=False, repr=False)

    def __post_init__(self) -> None:
        object.__setattr__(
            self,
            "_model",
            EllipticalSatelliteOrbitModel(
                semi_major_axis=self.semi_major_axis,
                inclination=self.inclination,
                raan=self.raan,
                phase_at_epoch=self.phase_at_epoch,
                epoch=self.epoch,
                satnum=self.satnum,
                gravity_model=self.gravity_model,
                drag_coeff=self.drag_coeff,
                eccentricity=self.eccentricity,
                argpo=self.argpo,
            ),
        )
        object.__setattr__(self, "_propagator", EllipticalSatelliteOrbitPropagator(self._model))

    @cached_property
    def ts(self) -> Timescale:
        return self._propagator.ts

    @cached_property
    def mean_motion(self) -> float:
        return self._propagator.mean_motion

    def datetime64_utc_to_skytime(self, t_datetime64: npt.NDArray[np.datetime64]) -> Time:
        return self._propagator.datetime64_utc_to_skytime(t_datetime64)

    @cached_property
    def satellite(self) -> EarthSatellite:
        return self._propagator.satellite

    def propagate(self, time: npt.NDArray[np.datetime64]) -> SatelliteOrbitState:
        return self._propagator.propagate(time)


def make_satellite_orbit(
    orbit_type: Literal["circular"],
    *args: Any,
    **kwargs: Any,
) -> SatelliteOrbit:
    if orbit_type == "circular":
        return CircularSatelliteOrbit(*args, **kwargs)
    else:
        msg = f"Unknown orbit type: {orbit_type}"
        raise ValueError(msg)



__all__ = [
    "Scenario",
]
from dataclasses import dataclass, field

import numpy as np
import numpy.typing as npt

# from cosmica.dynamics import SatelliteConstellation # This leads to circular import.
# Maybe we should move constellation to the `models` subpackage?
from .demand import Demand
from .gateway import Gateway
from .user import User


@dataclass(frozen=True, kw_only=True, slots=True)
class Scenario:
    """The simulation scenario."""

    time: npt.NDArray[np.datetime64]
    # constellation: SatelliteConstellation
    gateways: list[Gateway] = field(default_factory=list)
    users: list[User] = field(default_factory=list)
    demands: list[Demand] = field(default_factory=list)

    def build_topology(self) -> None:
        """Build the topology of the scenario."""
        # self.constellation.build_topology(self.gateways, self.users, self.demands)



__all__ = ["SatToSatBinaryCommLinkCalculatorWithRateCalc"]

from collections.abc import Collection
from itertools import chain
from typing import Annotated

import numpy as np
import numpy.typing as npt
from typing_extensions import Doc

from cosmica.dtos import DynamicsData
from cosmica.models import Satellite
from cosmica.utils.constants import BOLTZ_CONST, SPEED_OF_LIGHT
from cosmica.utils.vector import angle_between

from .base import CommLinkPerformance, MemorylessCommLinkCalculator


class SatToSatBinaryCommLinkCalculatorWithRateCalc(MemorylessCommLinkCalculator[Satellite, Satellite]):
    """Calculate satellite-to-satellite communication link performance for each edge in a network.

    The link performance is calculated as a binary value, i.e., 1 if the link is available and 0 otherwise.
    """

    def __init__(
        self,
        *,
        inter_satellite_link_capacity: float,
        max_inter_satellite_distance: float = float("inf"),
        lowest_altitude: float = 0.0,
        max_relative_angular_velocity: float = float("inf"),
        sun_exclusion_angle: float = 0.0,
        lct_power: float = 1.0,
        available_link_capacity: float = 1e9,
        lna_gain: float,
        noise_figure: float = 4,
    ) -> None:
        self.max_inter_satellite_distance = max_inter_satellite_distance
        self.inter_satellite_link_capacity = inter_satellite_link_capacity
        self.lowest_altitude = lowest_altitude
        self.max_relative_angular_velocity = max_relative_angular_velocity
        self.sun_exclusion_angle = sun_exclusion_angle
        self.lct_p0 = lct_power
        self.link_capacity = available_link_capacity
        self.lna_gain = lna_gain
        self.noise_figure = noise_figure

    def calc(
        self,
        edges: Collection[tuple[Satellite, Satellite]],
        *,
        dynamics_data: DynamicsData,
        rng: np.random.Generator,  # noqa: ARG002 For interface compatibility
    ) -> dict[tuple[Satellite, Satellite], CommLinkPerformance]:
        return {
            edge: self._calc_satellite_to_satellite(
                positions_eci=(
                    dynamics_data.satellite_position_eci[edge[0]],
                    dynamics_data.satellite_position_eci[edge[1]],
                ),
                velocities_eci=(
                    dynamics_data.satellite_velocity_eci[edge[0]],
                    dynamics_data.satellite_velocity_eci[edge[1]],
                ),
                attitude_angular_velocities_eci=(
                    dynamics_data.satellite_attitude_angular_velocity_eci[edge[0]],
                    dynamics_data.satellite_attitude_angular_velocity_eci[edge[1]],
                ),
                sun_direction_eci=dynamics_data.sun_direction_eci,
            )
            for edge in edges
        }

    def _calc_capacity_from_distance(
        self,
        distance: Annotated[
            float,
            Doc("Distance between satellites"),
        ],
    ) -> float:
        power = self.lct_p0 / distance**2  # can be replaced later by gaussian beam loss
        t_sys = 150  # K
        noise_factor = 10 ** (self.noise_figure / 10)
        noise = t_sys * self.link_capacity * BOLTZ_CONST * noise_factor
        gain = 10 ** (self.lna_gain / 10)  # convert from dB to W
        snr = gain * power / noise
        return self.link_capacity * np.log2(1 + snr)

    def _calc_satellite_to_satellite(
        self,
        *,
        positions_eci: Annotated[
            tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
            Doc("Position vectors in ECI frame. Shape: (3,)"),
        ],
        velocities_eci: Annotated[
            tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
            Doc("Velocity vectors in ECI frame. Shape: (3,)"),
        ],
        attitude_angular_velocities_eci: tuple[npt.NDArray[np.floating], npt.NDArray[np.floating]],
        sun_direction_eci: Annotated[
            npt.NDArray[np.floating],
            Doc("Sun direction vector in ECI frame. Shape: (3,)"),
        ],
    ) -> CommLinkPerformance:
        """Calculate binary communication link performance between two satellites."""
        for vec in chain(positions_eci, velocities_eci):
            assert vec.shape == (3,), f"Position and velocity vectors must be 3-dimensional, but got shape {vec.shape}"

        distance = float(np.linalg.norm(positions_eci[1] - positions_eci[0]))

        # Calculate relative angular velocity due to translational motion
        # The calculated angular velocity is that of the second satellite relative to the first satellite.
        relative_position_eci = positions_eci[1] - positions_eci[0]
        relative_velocity_eci = velocities_eci[1] - velocities_eci[0]
        relative_angular_velocity_translational_eci = (
            np.cross(relative_position_eci, relative_velocity_eci) / distance**2
        )

        relative_angular_velocities = (
            relative_angular_velocity_translational_eci - attitude_angular_velocities_eci[0],
            # Note: The relative angular velocity should be negated to get the angular velocity of the first satellite
            # relative to the second satellite.
            -relative_angular_velocity_translational_eci - attitude_angular_velocities_eci[1],
        )

        edge_sun_angle = angle_between(relative_position_eci, sun_direction_eci)

        link_available = bool(
            distance < self.max_inter_satellite_distance
            and all(
                float(np.linalg.norm(relative_angular_velocity)) < self.max_relative_angular_velocity
                for relative_angular_velocity in relative_angular_velocities
            )
            and edge_sun_angle >= self.sun_exclusion_angle
            and edge_sun_angle <= np.pi - self.sun_exclusion_angle,
        )

        return CommLinkPerformance(
            link_capacity=self._calc_capacity_from_distance(distance) if link_available else 0.0,
            delay=float(distance / SPEED_OF_LIGHT),
            link_available=link_available,
        )



__all__ = [
    "ConstantCommunicationDemand",
    "Demand",
    "OneTimeCommunicationDemand",
    "load_demands_from_toml_file",
    "parse_demand_item",
]
import tomllib
from abc import ABC, abstractmethod
from collections.abc import Hashable, Mapping
from dataclasses import dataclass
from datetime import UTC
from pathlib import Path
from typing import Any, Literal, Self

import numpy as np
from typing_extensions import deprecated

from .node import NodeGID


@dataclass(frozen=True, kw_only=True, slots=True)
class Demand[T: Hashable](ABC):
    """Base model for a demand."""

    id: T

    @classmethod
    @abstractmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_demand_item(cls, item: Mapping[str, Any]) -> Self:
        """Parse a demand item."""
        ...


@dataclass(frozen=True, kw_only=True, slots=True)
class ConstantCommunicationDemand[T: Hashable](Demand[T]):
    """Constant communication demand model."""

    id: T
    source: NodeGID
    destination: NodeGID
    distribution: Literal["uniform", "poisson"]
    transmission_rate: float

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_demand_item(cls, item: Mapping[str, Any]) -> Self:
        return cls(
            id=item["id"],
            source=item["src_gid"],
            destination=item["dst_gid"],
            transmission_rate=item["transmission_rate_bps"],
            distribution=item["distribution"],
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class TemporaryCommunicationDemand[T: Hashable](Demand[T]):
    """Temporary communication demand model."""

    id: T
    source: NodeGID
    destination: NodeGID
    transmission_rate: float
    distribution: Literal["uniform", "poisson"]
    start_time: np.datetime64
    end_time: np.datetime64

    def is_active(
        self,
        current_time: np.datetime64,
    ) -> np.bool:
        return self.start_time <= current_time < self.end_time

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_demand_item(cls, item: Mapping[str, Any]) -> Self:
        return cls(
            id=item["id"],
            source=item["src_gid"],
            destination=item["dst_gid"],
            transmission_rate=item["transmission_rate_bps"],
            distribution=item["distribution"],
            start_time=np.datetime64(item["start_time"].astimezone(tz=UTC).replace(tzinfo=None)),
            end_time=np.datetime64(item["end_time"].astimezone(tz=UTC).replace(tzinfo=None)),
        )


@dataclass(frozen=True, kw_only=True, slots=True)
class OneTimeCommunicationDemand[T: Hashable](Demand[T]):
    """One-time communication demand model.

    This model is used for the communication demand to transfer a certain amount of data
    from a source to a destination. The data is generated at the source at a certain
    time and the communication demand is created at the same time. The time by which the
    data transfer should be completed is also given.
    """

    id: T
    source: NodeGID
    destination: NodeGID
    data_size: float
    generation_time: np.datetime64
    deadline: np.datetime64

    @classmethod
    @deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
    def parse_demand_item(cls, item: Mapping[str, Any]) -> Self:
        return cls(
            id=item["id"],
            source=item["src_gid"],
            destination=item["dst_gid"],
            data_size=item["data_size"],
            generation_time=item["generation_time"],
            deadline=item["deadline"],
        )


_DEMAND_TYPES: dict[str, type[Demand[Hashable]]] = {
    "constant": ConstantCommunicationDemand,
    "temporary": TemporaryCommunicationDemand,
    "one_time": OneTimeCommunicationDemand,
}


@deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
def parse_demand_item(item: Mapping[str, Any]) -> Demand[Hashable]:
    """Parse a demand item."""
    demand_type = item["type"]
    return _DEMAND_TYPES[demand_type].parse_demand_item(item)


@deprecated("Construction of objects from TOML files is deprecated and will be removed in future versions.")
def load_demands_from_toml_file(toml_file_path: str | Path) -> list[Demand[Hashable]]:
    """Load demands from a TOML file."""
    toml_file_path = Path(toml_file_path)
    with toml_file_path.open("rb") as f:
        toml_data = tomllib.load(f)
    return list(map(parse_demand_item, toml_data["demands"]))



from __future__ import annotations

__all__ = [
    "get_sun_direction_eci",
]
from typing import TYPE_CHECKING, Annotated, overload

import numpy as np
from typing_extensions import Doc

from cosmica.utils.coordinates import juliandate
from cosmica.utils.vector import normalize

if TYPE_CHECKING:
    import numpy.typing as npt


@overload
def get_sun_direction_eci(time: npt.NDArray[np.datetime64]) -> npt.NDArray[np.float64]: ...


@overload
def get_sun_direction_eci(time: np.datetime64) -> npt.NDArray[np.float64]: ...


def get_sun_direction_eci(
    time: Annotated[
        np.datetime64 | npt.NDArray[np.datetime64],
        Doc("single element or array of UTC time values in np.datetime64 format."),
    ],
) -> Annotated[
    npt.NDArray[np.float64],
    Doc(
        "Corresponding Sun direction vector(s) (x,y,z) in eci."
        " If the input is a single element, the output is a 1D array."
        " If the input is an array, the output is a 2D array with the shape (len(time), 3).",
    ),
]:
    """Sun direction approximation function.

    The approximation considers an unperturbed motion of the Earth around the Sun with mean orbital elements
    that approximate the Sun's elliptic orbit wrt Earth around the year 2000.
    The algorithm is based on Oliver Montenbruck, Eberhard Grill; "Satellite Orbits: Models, Methods and Applications";
    Springer-Verlag Berlin. The alterations to the original model follow JAXA's Attitude Control Handbook
    JERG-2-510-HB001.
    """
    # tdt: Number of Julian Centuries. J2000 epoch -> 2451545.0 TT
    tdt = (juliandate(time) - 2451545.0) / 36525.0

    # m: mean anomaly
    m = np.deg2rad(357.5256 + 35999.045 * tdt)

    # right-ascension + argument of perigee = 282.94 deg
    # s: Sun's ecliptic longitude
    s = m + np.deg2rad(282.94 + (6892.0 * np.sin(m) + 72.0 * np.sin(2.0 * m) + 1250.09115 * tdt) / 3600.0 - 0.002652)

    # Obliquity of the ecliptic in radians
    epsilon_deg = 23.43929111
    epsilon = np.deg2rad(epsilon_deg)

    # Compute Sun Vector @ J2000. Shape: (3,) or (3, len(time))
    sun_vec_eci = np.array([np.cos(s), np.sin(s) * np.cos(epsilon), np.sin(s) * np.sin(epsilon)], dtype=np.float64)

    # normalize
    sun_vec_eci = sun_vec_eci.T  # shape: (3,) or (len(time), 3)

    # shape: (3,) or (len(time), 3)
    return normalize(sun_vec_eci, axis=-1)  # type: ignore[return-value]



__all__ = [
    "preserve_tick_params",
]
import logging
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any, Literal

from matplotlib.axes import Axes

logger = logging.getLogger(__name__)
_WHICH_LIST: tuple[Literal["major", "minor"], Literal["major", "minor"]] = ("major", "minor")


def _save_tick_params(ax: Axes) -> dict[str, dict[str, dict[str, Any]]]:
    return {
        "x": {which: ax.xaxis.get_tick_params(which=which) for which in _WHICH_LIST},
        "y": {which: ax.yaxis.get_tick_params(which=which) for which in _WHICH_LIST},
    }


def _restore_tick_params(params: dict[str, dict[str, dict[str, Any]]], ax: Axes) -> None:
    for which in _WHICH_LIST:
        ax.xaxis.set_tick_params(which=which, **params["x"][which])
        ax.yaxis.set_tick_params(which=which, **params["y"][which])


@contextmanager
def preserve_tick_params(ax: Axes) -> Generator[None, Any, None]:
    params = _save_tick_params(ax)
    try:
        yield
    finally:
        _restore_tick_params(params, ax)



__all__ = [
    "CircularSatelliteOrbitModel",
    "CommunicationTerminal",
    "ConstantCommunicationDemand",
    "ConstellationSatellite",
    "Demand",
    "EllipticalSatelliteOrbitModel",
    "Gateway",
    "GatewayOGS",
    "GravityModel",
    "Internet",
    "Node",
    "NodeGID",
    "OneTimeCommunicationDemand",
    "OpticalCommunicationTerminal",
    "RFCommunicationTerminal",
    "Satellite",
    "SatelliteTerminal",
    "Scenario",
    "StationaryOnGroundUser",
    "User",
    "UserOpticalCommunicationTerminal",
    "UserSatellite",
    "load_demands_from_toml_file",
    "load_users_from_toml_file",
]

from .demand import ConstantCommunicationDemand, Demand, OneTimeCommunicationDemand, load_demands_from_toml_file
from .gateway import Gateway, GatewayOGS
from .internet import Internet
from .node import Node, NodeGID
from .orbit import CircularSatelliteOrbitModel, EllipticalSatelliteOrbitModel, GravityModel
from .satellite import ConstellationSatellite, Satellite, SatelliteTerminal, UserSatellite
from .scenario import Scenario
from .terminal import (
    CommunicationTerminal,
    OpticalCommunicationTerminal,
    RFCommunicationTerminal,
    UserOpticalCommunicationTerminal,
)
from .user import StationaryOnGroundUser, User, load_users_from_toml_file



import logging
from functools import cached_property
from pathlib import Path

import numpy as np
import numpy.typing as npt
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.dates import DateFormatter
from matplotlib.ticker import AutoMinorLocator
from pandas import Series

from cosmica.dtos.dynamics_data import DynamicsData
from cosmica.experimental_packet_routing.dtos import (
    PacketRoutingResult,
    PacketRoutingSetting,
)
from cosmica.models.internet import Internet
from cosmica.models.satellite import ConstellationSatellite
from cosmica.visualization.equirectangular import (
    draw_countries,
    draw_lat_lon_grid,
    draw_snapshot,
)

logger = logging.getLogger(__name__)


# TODO (Takashima): 現状の描画は単一需要のみを対象としているため, 改善が必要
class RoutingResultVisualizer:
    def __init__(
        self,
        simulation_settings: PacketRoutingSetting,
        packet_routing_result: PacketRoutingResult,
    ) -> None:
        self.simulation_settings: PacketRoutingSetting = simulation_settings
        self.packet_routing_result: PacketRoutingResult = packet_routing_result

    def calculate_average_delay(
        self,
        *,
        time_from: np.datetime64 | None = None,
        time_to: np.datetime64 | None = None,
        weighted_data_size: bool = True,
    ) -> float:
        """到着したデータについて平均遅延時間を計算する."""
        calc_df = pd.DataFrame(
            {
                "generated_time": [
                    comm_data_demand.generated_time
                    for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "delay": [
                    comm_data_demand.delay for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "data_size": [
                    comm_data_demand.data_size for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
            },
        )

        # 時間範囲を指定してフィルタリング
        if time_from is not None:
            calc_df = calc_df[calc_df["generated_time"] >= time_from]
        if time_to is not None:
            calc_df = calc_df[calc_df["generated_time"] <= time_to]

        # 有限な遅延値のみを選択
        finite_delay_ser: Series = calc_df["delay"][np.isfinite(calc_df["delay"])]
        finite_data_size_ser: pd.Series = calc_df["data_size"][np.isfinite(calc_df["delay"])]

        if weighted_data_size:
            # データサイズで重み付けした平均遅延時間を計算
            average_delay: float = np.average(finite_delay_ser, weights=finite_data_size_ser)
        else:
            # 単純な平均遅延時間を計算
            average_delay = finite_delay_ser.mean()

        logger.info(f"Average delay: {average_delay} s")

        return average_delay

    def calculate_average_increased_delay(
        self,
        time_baseline: np.datetime64,
        *,
        time_from: np.datetime64 | None = None,
        time_to: np.datetime64 | None = None,
        weighted_data_size: bool = True,
    ) -> float:
        """到着したデータについて平均遅延時間の、ある時刻の遅延時間に対する増加量を計算する."""
        calc_df = pd.DataFrame(
            {
                "generated_time": [
                    comm_data_demand.generated_time
                    for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "delay": [
                    comm_data_demand.delay for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "data_size": [
                    comm_data_demand.data_size for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
            },
        )

        # 時間範囲を指定してフィルタリング
        if time_from is not None:
            calc_df = calc_df[calc_df["generated_time"] >= time_from]
        if time_to is not None:
            calc_df = calc_df[calc_df["generated_time"] <= time_to]

        # 有限な遅延値のみを選択
        finite_delay_ser: Series = calc_df["delay"][np.isfinite(calc_df["delay"])]
        finite_data_size_ser: pd.Series = calc_df["data_size"][np.isfinite(calc_df["delay"])]

        if weighted_data_size:
            # データサイズで重み付けした平均遅延時間を計算
            average_delay: float = np.average(
                finite_delay_ser,
                weights=finite_data_size_ser,
            )
        else:
            # 単純な平均遅延時間を計算
            average_delay = finite_delay_ser.mean()

        # 指定した時刻の遅延時間からの差分を計算
        closest_idx = (calc_df["generated_time"] - pd.to_datetime(time_baseline)).abs().idxmin()
        average_increased_delay = average_delay - pd.to_numeric(
            calc_df.loc[closest_idx, "delay"],
        )

        logger.info(f"Average increased delay: {average_increased_delay} s")

        return average_delay

    def calculate_average_packet_loss_rate(
        self,
        *,
        time_from: np.datetime64 | None = None,
        time_to: np.datetime64 | None = None,
        weighted_data_size: bool = True,
    ) -> float:
        """生成データの平均パケットロス率を計算する."""
        calc_df = pd.DataFrame(
            {
                "generated_time": [
                    comm_data_demand.generated_time
                    for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "reach_dst": [
                    comm_data_demand.reach_dst for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "data_size": [
                    comm_data_demand.data_size for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
            },
        )

        # 時間範囲を指定してフィルタリング
        if time_from is not None:
            calc_df = calc_df[calc_df["generated_time"] >= time_from]
        if time_to is not None:
            calc_df = calc_df[calc_df["generated_time"] <= time_to]

        if weighted_data_size:
            # データサイズで重み付けした平均パケットロス率を計算
            average_packet_reach_rate: float = np.average(calc_df["reach_dst"], weights=calc_df["data_size"])
        else:
            average_packet_reach_rate = calc_df["reach_dst"].mean()

        average_packet_loss_rate = 1 - average_packet_reach_rate
        logger.info(f"Average packet loss rate: {average_packet_loss_rate}")

        return average_packet_loss_rate

    def plot_delay(
        self,
        save_path: Path,
        *,
        with_title: bool = False,
        title_fontsize: int = 16,
        label_fontsize: int = 14,
        legend_fontsize: int = 12,
        tick_label_fontsize: int = 12,
        legend_loc: str = "best",
        xlim: tuple | None = None,
        ylim: tuple | None = None,
        dpi: int = 300,
        with_grid: bool = False,
        use_time_from_epoch: bool = False,
        replace_inf: bool = False,
        # link_failure_timing=None,
    ) -> None:
        ## Convert comm_data_demand_list into a pandas DataFrame
        # 到達したデータ, もしくはパケットロスしたデータのみプロットする
        plot_df = pd.DataFrame(
            {
                "generated_time": [
                    comm_data_demand.generated_time
                    for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "delay": [
                    comm_data_demand.delay for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
            },
        )
        plot_df = plot_df.sort_values("generated_time")

        if use_time_from_epoch:
            x_data = (plot_df["generated_time"] - self.time[0]) / np.timedelta64(1, "s")
        else:
            x_data = plot_df["generated_time"]

        max_delay = plot_df["delay"][plot_df["delay"] != np.inf].max()

        if replace_inf:
            # np.infだとプロットされないので、適当な大きい値に置換
            plot_df["delay"] = plot_df["delay"].replace(np.inf, 1e5)

        ## プロット
        fig, ax = plt.subplots(figsize=(10, 6), dpi=dpi)

        ax.plot(x_data, plot_df["delay"], label="Delay", color="blue")

        # Customizing title and labels font size
        if with_title:
            ax.set_title("Delay", fontsize=title_fontsize)
        ax.set_ylabel("Average delay time [s]", fontsize=label_fontsize)
        if use_time_from_epoch:
            ax.set_xlabel("Time from epoch[s]", fontsize=label_fontsize)
        else:
            ax.set_xlabel("Time", fontsize=label_fontsize)
            # Formatting the x-axis with date and time, and rotating the text
            ax.xaxis.set_major_formatter(DateFormatter("%Y/%m/%d %H:%M:%S"))
            plt.setp(ax.get_xticklabels(), rotation=45, ha="right")

        # Customizing tick label font size
        ax.tick_params(axis="both", labelsize=tick_label_fontsize)

        # 軸の範囲設定
        if use_time_from_epoch:
            ax.set_xlim(xlim if xlim else (self.time_from_epoch[0], self.time_from_epoch[-1]))
        else:
            ax.set_xlim(xlim if xlim else (self.time[0], self.time[-1]))

        if max_delay == np.inf or np.isnan(max_delay):
            ax.set_ylim(0, None)
        else:
            ax.set_ylim(ylim if ylim else (0, max_delay * 1.1))

        if with_grid:
            ax.xaxis.set_minor_locator(AutoMinorLocator(5))
            ax.yaxis.set_minor_locator(AutoMinorLocator(5))
            ax.grid(which="major", color="#CCCCCC", linestyle="--")
            ax.grid(which="minor", color="#CCCCCC", linestyle=":")

        # 凡例の設定
        ax.legend(fontsize=legend_fontsize, loc=legend_loc)

        # 保存
        fig.savefig(save_path, bbox_inches="tight")
        logger.info(f"Saved visualization: {save_path}")

    def plot_arrival_data_rate(
        self,
        save_path: Path,
        *,
        with_title: bool = False,
        title_fontsize: int = 16,
        label_fontsize: int = 14,
        legend_fontsize: int = 12,
        tick_label_fontsize: int = 12,
        legend_loc: str = "best",
        xlim: tuple | None = None,
        ylim: tuple | None = None,
        dpi: int = 300,
        with_grid: bool = False,
        use_time_from_epoch: bool = False,
        # link_failure_timing=None,
    ) -> None:
        # --- 1. 必要データを整形 ---------------------------------------------------
        plot_df = pd.DataFrame(
            {
                "generated_time": [
                    comm_data_demand.generated_time
                    for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "delay": [
                    comm_data_demand.delay for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "data_size": [
                    comm_data_demand.data_size for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "reach_dst": [
                    comm_data_demand.reach_dst for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
            },
        )
        plot_df = plot_df[plot_df["reach_dst"]]
        plot_df["arrival_time"] = (plot_df["generated_time"] + pd.to_timedelta(plot_df["delay"], unit="s")).astype(
            "datetime64[ms]",
        )

        # --- 2. タイムステップ情報 --------------------------------------------------
        time_index = pd.Series(self.time).astype("datetime64[ms]")
        dt_seconds = np.asarray(self.time_step_array, dtype=float)

        # --- 3. 各 arrival を直前の time_index に割り当て ---------------------------
        # np.searchsorted で挿入位置を取得 → 1 つ前のインデックスが「属するステップ」
        idx = np.searchsorted(time_index.to_numpy(), plot_df["arrival_time"].to_numpy(), side="right") - 1
        valid = idx >= 0  # 負インデックスは epoch 以前なので除外

        # --- 4. ステップ毎にバイト数を加算 -----------------------------------------
        bytes_per_step = np.zeros(len(time_index), dtype=float)
        np.add.at(bytes_per_step, idx[valid], plot_df["data_size"].to_numpy()[valid])

        arrival_rate = bytes_per_step / dt_seconds

        # --- 5. 可視化 -------------------------------------------------------------
        x_data = (time_index - time_index.iloc[0]).dt.total_seconds() if use_time_from_epoch else time_index

        fig, ax = plt.subplots(figsize=(10, 6), dpi=dpi)
        ax.plot(x_data, arrival_rate, label="Arrival data rate", lw=1)

        if with_title:
            ax.set_title("Arrival Data Rate", fontsize=title_fontsize)
        ax.set_ylabel("Arrival data rate [bit/s]", fontsize=label_fontsize)
        if use_time_from_epoch:
            ax.set_xlabel("Time from Epoch [s]", fontsize=label_fontsize)
        else:
            ax.set_xlabel("Time", fontsize=label_fontsize)
            ax.xaxis.set_major_formatter(DateFormatter("%Y/%m/%d %H:%M:%S"))
            plt.setp(ax.get_xticklabels(), rotation=45, ha="right")

        ax.tick_params(axis="both", labelsize=tick_label_fontsize)
        ax.set_xlim(xlim if xlim else (x_data.iloc[0], x_data.iloc[-1]))
        ax.set_ylim(ylim if ylim else (0, None))

        if with_grid:
            ax.xaxis.set_minor_locator(AutoMinorLocator(5))
            ax.yaxis.set_minor_locator(AutoMinorLocator(5))
            ax.grid(which="major", linestyle="--", alpha=0.6)
            ax.grid(which="minor", linestyle=":", alpha=0.4)

        ax.legend(fontsize=legend_fontsize, loc=legend_loc)
        fig.savefig(save_path, bbox_inches="tight")
        logger.info(f"Saved visualization: {save_path}")

    def plot_accumulated_arrival_data_size(
        self,
        save_path: Path,
        *,
        with_title: bool = False,
        title_fontsize: int = 16,
        label_fontsize: int = 14,
        legend_fontsize: int = 12,
        tick_label_fontsize: int = 12,
        legend_loc: str = "best",
        xlim: tuple | None = None,
        ylim: tuple | None = None,
        dpi: int = 300,
        with_grid: bool = False,
        use_time_from_epoch: bool = False,
        # link_failure_timing=None,
    ) -> None:
        ## Convert comm_data_demand_list into a pandas DataFrame
        # 到達したデータ, もしくはパケットロスしたデータのみプロットする
        plot_df = pd.DataFrame(
            {
                "generated_time": [
                    comm_data_demand.generated_time
                    for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "delay": [
                    comm_data_demand.delay for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "data_size": [
                    comm_data_demand.data_size for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
                "reach_dst": [
                    comm_data_demand.reach_dst for comm_data_demand in self.packet_routing_result.comm_data_demand_list
                ],
            },
        )
        plot_df = plot_df[plot_df["reach_dst"]]

        plot_df["arrival_time"] = plot_df["generated_time"] + pd.to_timedelta(plot_df["delay"], unit="s")
        plot_df = plot_df.sort_values("arrival_time")

        plot_df["accumulated_arrival_data_size"] = plot_df.groupby("reach_dst")["data_size"].cumsum()

        if use_time_from_epoch:
            x_data = (plot_df["arrival_time"] - self.time[0]) / np.timedelta64(1, "s")
        else:
            x_data = plot_df["arrival_time"]

        ## プロット
        fig, ax = plt.subplots(figsize=(10, 6), dpi=dpi)

        ax.plot(
            x_data,
            plot_df["accumulated_arrival_data_size"],
            label="Accumulated arrival data size",
            color="blue",
        )

        # Customizing title and labels font size
        if with_title:
            ax.set_title("Accumulated Arrival Data Size", fontsize=title_fontsize)
        ax.set_ylabel("Accumulated arrival data size [bit]", fontsize=label_fontsize)
        if use_time_from_epoch:
            ax.set_xlabel("Time from Epoch [s]", fontsize=label_fontsize)
        else:
            ax.set_xlabel("Time", fontsize=label_fontsize)
            # Formatting the x-axis with date and time, and rotating the text
            ax.xaxis.set_major_formatter(DateFormatter("%Y/%m/%d %H:%M:%S"))
            plt.setp(ax.get_xticklabels(), rotation=45, ha="right")

        # Customizing tick label font size
        ax.tick_params(axis="both", labelsize=tick_label_fontsize)

        # 軸の範囲設定
        if use_time_from_epoch:
            ax.set_xlim(xlim if xlim else (self.time_from_epoch[0], self.time_from_epoch[-1]))
        else:
            ax.set_xlim(xlim if xlim else (self.time[0], self.time[-1]))
        ax.set_ylim(ylim if ylim else (0, None))

        if with_grid:
            ax.xaxis.set_minor_locator(AutoMinorLocator(5))
            ax.yaxis.set_minor_locator(AutoMinorLocator(5))
            ax.grid(which="major", color="#CCCCCC", linestyle="--")
            ax.grid(which="minor", color="#CCCCCC", linestyle=":")

        # 凡例の設定
        ax.legend(fontsize=legend_fontsize, loc=legend_loc)

        # 保存
        fig.savefig(save_path, bbox_inches="tight")
        logger.info(f"Saved visualization: {save_path}")

    def plot_graph_at_certain_time(
        self,
        time: np.datetime64,
        dynamics_data: DynamicsData[ConstellationSatellite],
        save_path: Path,
        *,
        with_demand_data: bool = True,
        with_lsa_data: bool = False,
        dpi: int = 100,
    ) -> None:
        """指定した時間のグラフを描画する."""
        time_indices = np.where(self.time == time)[0]
        if len(time_indices) == 0:
            msg = f"Time {time} not found in self.time array."
            raise ValueError(msg)
        time_index = time_indices[0]

        edges_with_demand_data = self._get_edges_with_demand_data(time_index) if with_demand_data else set()

        edges_with_lsa_data = self._get_edges_with_lsa_data(time_index) if with_lsa_data else set()
        # routing_table_arrows = get_routing_table_arrows(time_idx)

        fig, ax = plt.subplots(figsize=(15, 8))

        title = f"Time: {np.datetime_as_string(time, unit='ms').split('T')[1]}"
        ax.set_title(title)

        draw_lat_lon_grid(ax=ax)
        draw_countries(ax=ax)
        draw_snapshot(
            graph=self.packet_routing_result.all_graphs_after_simulation[time_index],
            dynamics_data=dynamics_data[time_index],
            ax=ax,
            with_labels=False,
            focus_edges_list=[edges_with_demand_data, edges_with_lsa_data],
            focus_edges_label_list=["Demand data", "LSA data"],
        )

        ax.legend(loc="lower left")

        # 保存
        fig.savefig(save_path, bbox_inches="tight", dpi=dpi)
        logger.info(f"Saved visualization: {save_path}")

    def plot_graph_over_time(
        self,
        time_index_from: int,
        time_index_to: int,
        time_index_step: int,
        dynamics_data: DynamicsData[ConstellationSatellite],
        save_path: Path,
        *,
        with_demand_data: bool = True,
        with_lsa_data: bool = False,
        dpi: int = 100,
    ) -> None:
        """時間経過に伴うグラフの変化を描画する."""
        # TODO(Takashima): time_index でなく timeを引数に取れるようにする
        n_fig = (time_index_to - time_index_from) // time_index_step
        ncols = 3
        nrows = (n_fig + ncols - 1) // ncols

        fig, axes = plt.subplots(
            nrows=nrows,
            ncols=ncols,
            figsize=(15 * ncols, 8 * nrows),
        )
        fig.subplots_adjust(wspace=0.1, hspace=0.1)

        for idx, time_index in enumerate(range(time_index_from, time_index_to, time_index_step)):
            row, col = divmod(idx, ncols)

            edges_with_demand_data = self._get_edges_with_demand_data(time_index) if with_demand_data else set()

            edges_with_lsa_data = self._get_edges_with_lsa_data(time_index) if with_lsa_data else set()

            title = f"Time: {np.datetime_as_string(self.time[time_index], unit='ms').split('T')[1]}"
            axes[row, col].set_title(title)

            draw_lat_lon_grid(ax=axes[row, col])
            draw_countries(ax=axes[row, col])
            draw_snapshot(
                graph=self.packet_routing_result.all_graphs_after_simulation[time_index],
                dynamics_data=dynamics_data[time_index],
                ax=axes[row, col],
                with_labels=False,
                focus_edges_list=[edges_with_demand_data, edges_with_lsa_data],
                focus_edges_label_list=["Demand data", "LSA data"],
            )

        # 保存
        fig.savefig(save_path, bbox_inches="tight", dpi=dpi)
        logger.info(f"Saved visualization: {save_path}")

    def plot_graph_animation(
        self,
        time_index_from: int,
        time_index_to: int,
        time_index_step: int,
        dynamics_data: DynamicsData[ConstellationSatellite],
        save_path: Path,
        *,
        with_demand_data: bool = True,
        with_lsa_data: bool = False,
        dpi: int = 100,
    ) -> None:
        """時間経過に伴うグラフの変化をアニメーションで描画する."""
        # TODO(Takashima): time_index でなく timeを引数に取れるようにする
        time_index_for_plot = range(time_index_from, time_index_to, time_index_step)

        fig, ax = plt.subplots(figsize=(15, 8))

        def update(frame: int):  # noqa: ANN202
            ax.clear()
            time_index = time_index_for_plot[frame]

            edges_with_demand_data = self._get_edges_with_demand_data(time_index) if with_demand_data else set()

            edges_with_lsa_data = self._get_edges_with_lsa_data(time_index) if with_lsa_data else set()

            title = f"Time: {np.datetime_as_string(self.time[time_index], unit='ms').split('T')[1]}"
            ax.set_title(title)

            draw_lat_lon_grid(ax=ax)
            draw_countries(ax=ax)
            draw_snapshot(
                graph=self.packet_routing_result.all_graphs_after_simulation[time_index],
                dynamics_data=dynamics_data[time_index],
                ax=ax,
                with_labels=False,
                focus_edges_list=[edges_with_demand_data, edges_with_lsa_data],
                focus_edges_label_list=["Demand data", "LSA data"],
            )
            ax.legend(loc="lower left")

        ani = FuncAnimation(fig, update, frames=len(time_index_for_plot), interval=100)
        ani.save(
            save_path,
            dpi=dpi,
        )

    def _get_edges_with_demand_data(
        self,
        time_idx: int,
    ) -> set:
        edges_with_data: set = {
            edge
            for edge in self.packet_routing_result.all_graphs_after_simulation[time_idx].edges
            if self.packet_routing_result.all_graphs_after_simulation[time_idx].edges[edge][
                "bandwidth_usage_for_demand_data"
            ]
            > 0
        }
        return {edge for edge in edges_with_data if all(not isinstance(node, Internet) for node in edge)}

    def _get_edges_with_lsa_data(
        self,
        time_idx: int,
    ) -> set:
        edges_with_data: set = {
            edge
            for edge in self.packet_routing_result.all_graphs_after_simulation[time_idx].edges
            if self.packet_routing_result.all_graphs_after_simulation[time_idx].edges[edge][
                "bandwidth_usage_for_lsa_data"
            ]
            > 0
        }
        return {edge for edge in edges_with_data if all(not isinstance(node, Internet) for node in edge)}

    # def plot_bandwidth_on_edges(
    #     self,
    #     save_path: Path,
    #     *,
    #     title_fontsize: int = 16,
    #     label_fontsize: int = 14,
    #     legend_fontsize: int = 12,
    #     tick_label_fontsize: int = 12,
    #     legend_loc: str = "best",
    #     xlim: tuple | None = None,
    #     ylim: tuple | None = None,
    #     dpi: int = 300,
    #     with_grid: bool = False,
    #     use_time_from_epoch: bool = False,
    #     save_name: str = "bandwidth_on_edges",
    #     # link_failure_timing=None,
    # ) -> None:
    #     def get_bandwidth_usage_on_edge(
    #         graphs: list[nx.Graph],
    #         edge_name: tuple[Any, Any],
    #         time_array: npt.NDArray,
    #     ) -> ndarray[float]:
    #         """指定したエッジの帯域幅使用量を取得する関数."""
    #         bandwidth_usage = np.zeros(len(time_array))
    #         for time_idx in range(len(time_array)):
    #             if graphs[time_idx].has_edge(*edge_name):
    #                 bandwidth_usage[time_idx] = graphs[time_idx].edges[edge_name]["bandwidth_usage_for_demand"]
    #         return bandwidth_usage

    #     ## Convert comm_data_demand_list into a pandas DataFrame
    #     # 到達したデータ, もしくはパケットロスしたデータのみプロットする
    #     plot_df = pd.DataFrame(
    #         {
    #             "generated_time":
    # [comm_data_demand.generated_time for comm_data_demand in self.comm_data_demand_list],
    #             "delay": [comm_data_demand.delay for comm_data_demand in self.comm_data_demand_list],
    #             "data_size": [comm_data_demand.data_size for comm_data_demand in self.comm_data_demand_list],
    #             "reach_dst": [comm_data_demand.reach_dst for comm_data_demand in self.comm_data_demand_list],
    #         },
    #     )
    #     plot_df["arrival_time"] = plot_df["generated_time"] + pd.to_timedelta(plot_df["delay"], unit="s")
    #     plot_df: DataFrame = plot_df.sort_values("arrival_time")

    #     plot_df["accumulated_arrival_data_size"] = plot_df.groupby("reach_dst")["data_size"].cumsum()

    #     if use_time_from_epoch:
    #         x_data = (plot_df["arrival_time"] - self.time[0]) / np.timedelta64(1, "s")
    #     else:
    #         x_data = plot_df["arrival_time"]

    #     # 各エッジに対して帯域幅使用量を取得
    #     for graph_idx, _graph in enumerate(self.all_graphs_after_simulation)):
    #         for edge in _graph.edges:
    #             bandwidth_usage_on_edge =
    # get_bandwidth_usage_on_edge(self.all_graphs_after_simulation, edge, time_array)
    #             if np.any(bandwidth_usage_on_edge >= 0.8e9):  # 帯域幅使用量が閾値を超えた場合のみプロット
    #                 edge_label = ", ".join([key for key, value in nodes_dict.items() if value in (edge[0], edge[1])])
    #                 ax.plot(time_array_from_start_time, bandwidth_usage_on_edge, label=edge_label)

    #     ## プロット
    #     fig, ax = plt.subplots(figsize=(10, 6), dpi=dpi)

    #     ax.plot(x_data, plot_df["accumulated_arrival_data_size"], label="Accumulated arrival data size", color="blue")

    #     # Customizing title and labels font size
    #     ax.set_title("Accumulated Arrival Data Size", fontsize=title_fontsize)
    #     ax.set_ylabel("Accumulated arrival data size [bit]", fontsize=label_fontsize)
    #     if use_time_from_epoch:
    #         ax.set_xlabel("Time from Epoch [s]", fontsize=label_fontsize)
    #     else:
    #         ax.set_xlabel("Time", fontsize=label_fontsize)
    #         # Formatting the x-axis with date and time, and rotating the text
    #         ax.xaxis.set_major_formatter(DateFormatter("%Y/%m/%d %H:%M:%S"))
    #         plt.setp(ax.get_xticklabels(), rotation=45, ha="right")

    #     # Customizing tick label font size
    #     ax.tick_params(axis="both", labelsize=tick_label_fontsize)

    #     # 軸の範囲設定
    #     if use_time_from_epoch:
    #         ax.set_xlim(xlim if xlim else (self.time_from_epoch[0], self.time_from_epoch[-1]))
    #     else:
    #         ax.set_xlim(xlim if xlim else (self.time[0], self.time[-1]))
    #     ax.set_ylim(ylim if ylim else (0, None))

    #     if with_grid:
    #         ax.xaxis.set_minor_locator(AutoMinorLocator(5))
    #         ax.yaxis.set_minor_locator(AutoMinorLocator(5))
    #         ax.grid(which="major", color="#CCCCCC", linestyle="--")
    #         ax.grid(which="minor", color="#CCCCCC", linestyle=":")

    #     # 凡例の設定
    #     ax.legend(fontsize=legend_fontsize, loc=legend_loc)

    #     # 保存
    #     save_path: Path = save_path / (save_name + ".png")
    #     fig.savefig(save_path, bbox_inches="tight")
    #     logging.info(f"Saved visualization: {save_path}")

    @cached_property
    def time(self) -> npt.NDArray:
        return self.simulation_settings.time

    @cached_property
    def time_from_epoch(self) -> npt.NDArray:
        return (self.simulation_settings.time - self.simulation_settings.time[0]) / np.timedelta64(1, "s")

    @cached_property
    def time_step_array(self) -> npt.NDArray:
        time_from_epoch_diff = np.diff(self.time_from_epoch)
        return np.append(time_from_epoch_diff, time_from_epoch_diff[-1])



__all__ = [
    "GatewayToGatewayCommLinkCalculator",
]

from collections.abc import Collection

import numpy as np

from cosmica.dtos import DynamicsData
from cosmica.models import Gateway
from cosmica.utils.constants import EARTH_RADIUS, SPEED_OF_LIGHT

from .base import CommLinkPerformance, MemorylessCommLinkCalculator


class GatewayToGatewayCommLinkCalculator(MemorylessCommLinkCalculator[Gateway, Gateway]):
    """Calculate gateway-to-gateway communication link performance using great circle distance."""

    def __init__(
        self,
        *,
        gateway_to_gateway_bandwidth: float,
        refractive_index: float = 1.5,
    ) -> None:
        """Initialize the gateway-to-gateway communication link calculator."""
        self.gateway_to_gateway_bandwidth = gateway_to_gateway_bandwidth
        self.refractive_index = refractive_index

    def calc(
        self,
        edges: Collection[tuple[Gateway, Gateway]],
        *,
        dynamics_data: DynamicsData,  # noqa: ARG002 For interface compatibility
        rng: np.random.Generator,  # noqa: ARG002 For interface compatibility
    ) -> dict[tuple[Gateway, Gateway], CommLinkPerformance]:
        """Calculate communication link performance for gateway-to-gateway edges."""
        return {
            edge: self._calc_gateway_to_gateway(
                gateway1=edge[0],
                gateway2=edge[1],
            )
            for edge in edges
        }

    def _calc_gateway_to_gateway(
        self,
        gateway1: Gateway,
        gateway2: Gateway,
    ) -> CommLinkPerformance:
        """Calculate communication link performance between two gateways."""
        # Calculate great circle distance between gateways
        distance = self._calc_great_circle_distance(
            lat1=gateway1.latitude,
            lon1=gateway1.longitude,
            lat2=gateway2.latitude,
            lon2=gateway2.longitude,
        )

        # Calculate propagation delay with refractive index
        propagation_speed = SPEED_OF_LIGHT / self.refractive_index
        delay = distance / propagation_speed

        # Gateway-to-gateway links are always available
        link_available = True

        return CommLinkPerformance(
            link_capacity=self.gateway_to_gateway_bandwidth if link_available else 0.0,
            delay=float(delay),
            link_available=link_available,
        )

    @staticmethod
    def _calc_great_circle_distance(
        lat1: float,
        lon1: float,
        lat2: float,
        lon2: float,
    ) -> float:
        """Calculate the great circle distance between two points on Earth's surface using the haversine formula."""
        # Haversine formula
        dlat = lat2 - lat1
        dlon = lon2 - lon1

        a = np.sin(dlat / 2) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2) ** 2
        c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))

        return EARTH_RADIUS * c



__all__ = [
    "ApertureAveragedLogNormalScintillationModel",
    "AtmosphericScintillationModel",
    "BinaryCloudModel",
    "CloudStates",
    "CommLinkCalculationCoordinator",
    "CommLinkCalculator",
    "CommLinkPerformance",
    "EdgeFailureModel",
    "ExpEdgeModel",
    "GatewayToGatewayCommLinkCalculator",
    "GatewayToInternetCommLinkCalculator",
    "GeometricCommLinkCalculator",
    "MemorylessCommLinkCalculator",
    "MemorylessCommLinkCalculatorWrapper",
    "OTC2OTCBinaryCommLinkCalculator",
    "SatToGatewayBinaryCommLinkCalculator",
    "SatToGatewayBinaryCommLinkCalculatorWithScintillation",
    "SatToGatewayBinaryMemoryCommLinkCalculator",
    "SatToGatewayStochasticBinaryCommLinkCalculator",
    "SatToSatBinaryCommLinkCalculator",
    "SatToSatBinaryCommLinkCalculatorWithRateCalc",
    "SatToSatBinaryMemoryCommLinkCalculator",
]
from .base import (
    CommLinkCalculator,
    CommLinkPerformance,
    MemorylessCommLinkCalculator,
    MemorylessCommLinkCalculatorWrapper,
)
from .coordinator import CommLinkCalculationCoordinator
from .gateway_to_gateway import GatewayToGatewayCommLinkCalculator
from .gateway_to_internet import GatewayToInternetCommLinkCalculator
from .geometric import GeometricCommLinkCalculator
from .rate_distance_calculator import SatToSatBinaryCommLinkCalculatorWithRateCalc
from .sat_to_ground import (
    SatToGatewayBinaryCommLinkCalculator,
    SatToGatewayBinaryCommLinkCalculatorWithScintillation,
    SatToGatewayBinaryMemoryCommLinkCalculator,
    SatToGatewayStochasticBinaryCommLinkCalculator,
)
from .sat_to_sat import (
    OTC2OTCBinaryCommLinkCalculator,
    SatToSatBinaryCommLinkCalculator,
    SatToSatBinaryMemoryCommLinkCalculator,
)
from .uncertainty import (
    ApertureAveragedLogNormalScintillationModel,
    AtmosphericScintillationModel,
    BinaryCloudModel,
    CloudStates,
    EdgeFailureModel,
    ExpEdgeModel,
)
